# Copyright (c) 2017-2025 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation
#     Jonathan Slenders - proposed changes to allow SFTP server callbacks
#                         to be coroutines

"""SCP handlers"""

import argparse
import asyncio
import inspect
from pathlib import PurePath
import posixpath
import shlex
import string
import sys
from types import TracebackType
from typing import TYPE_CHECKING, AsyncIterator, List, NoReturn, Optional
from typing import Sequence, Tuple, Type, Union, cast
from typing_extensions import Protocol, Self

from .constants import DEFAULT_LANG
from .constants import FILEXFER_TYPE_REGULAR, FILEXFER_TYPE_DIRECTORY
from .logging import SSHLogger
from .misc import BytesOrStr, FilePath, HostPort, MaybeAwait
from .misc import async_context_manager, plural
from .sftp import SFTPAttrs, SFTPGlob, SFTPName, SFTPServer, SFTPServerFS
from .sftp import SFTPError, SFTPFailure, SFTPBadMessage, SFTPConnectionLost
from .sftp import SFTPErrorHandler, SFTPProgressHandler, local_fs


if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .channel import SSHServerChannel
    from .connection import SSHClientConnection
    from .stream import SSHReader, SSHWriter


_SCPConn = Union[None, bytes, str, HostPort, 'SSHClientConnection']
_SCPPath = Union[bytes, FilePath]
_SCPConnPath = Union[Tuple[_SCPConn, _SCPPath], _SCPConn, _SCPPath]


_SCP_BLOCK_SIZE = 256*1024    # 256 KiB


class _SCPFileProtocol(Protocol):
    """Protocol for accessing a file during an SCP copy"""

    async def __aenter__(self) -> Self:
        """Allow _SCPFileProtocol to be used as an async context manager"""

    async def __aexit__(self, _exc_type: Optional[Type[BaseException]],
                        _exc_value: Optional[BaseException],
                        _traceback: Optional[TracebackType]) -> bool:
        """Wait for file close when used as an async context manager"""

    async def read(self, size: int, offset: int) -> bytes:
        """Read data from the local file"""

    async def write(self, data: bytes, offset: int) -> int:
        """Write data to the local file"""

    async def close(self) -> None:
        """Close the local file"""


class _SCPFSProtocol(Protocol):
    """Protocol for accessing a filesystem during an SCP copy"""

    @staticmethod
    def basename(path: bytes) -> bytes:
        """Return the final component of a POSIX-style path"""

    async def stat(self, path: bytes) -> 'SFTPAttrs':
        """Get attributes of a file or directory, following symlinks"""

    async def setstat(self, path: bytes, attrs: 'SFTPAttrs') -> None:
        """Set attributes of a file or directory"""

    async def exists(self, path: bytes) -> bool:
        """Return if a path exists"""

    async def isdir(self, path: bytes) -> bool:
        """Return if the path refers to a directory"""

    def scandir(self, path: bytes) -> AsyncIterator[SFTPName]:
        """Read the names and attributes of files in a directory"""

    async def mkdir(self, path: bytes) -> None:
        """Create a directory"""

    @async_context_manager
    async def open(self, path: bytes, mode: str) -> _SCPFileProtocol:
        """Open a file"""


def _scp_error(exc_class: Type[Exception], reason: BytesOrStr,
               path: Optional[bytes] = None, fatal: bool = False,
               suppress_send: bool = False,
               lang: str = DEFAULT_LANG) -> Exception:
    """Construct SCP version of SFTPError exception"""

    if isinstance(reason, bytes):
        reason = reason.decode('utf-8', errors='replace')

    if path:
        reason = reason + ': ' + path.decode('utf-8', errors='replace')

    exc = exc_class(reason, lang)

    setattr(exc, 'fatal', fatal)
    setattr(exc, 'suppress_send', suppress_send)

    return exc


def _parse_cd_args(args: bytes) -> Tuple[int, int, bytes]:
    """Parse arguments to an SCP copy or dir request"""

    try:
        permissions, size, name = args.split(None, 2)
        return int(permissions, 8), int(size), name
    except ValueError:
        raise _scp_error(SFTPBadMessage,
                         'Invalid copy or dir request') from None


def _parse_t_args(args: bytes) -> Tuple[int, int]:
    """Parse argument to an SCP time request"""

    try:
        mtime, _, atime, _ = args.split()
        return int(atime), int(mtime)
    except ValueError:
        raise _scp_error(SFTPBadMessage, 'Invalid time request') from None


async def _parse_path(path: _SCPConnPath, **kwargs) -> \
        Tuple[Optional['SSHClientConnection'], _SCPPath, bool]:
    """Convert an SCP path into an SSHClientConnection and path"""

    # pylint: disable=cyclic-import,import-outside-toplevel
    from . import connect

    conn: _SCPConn

    if isinstance(path, tuple):
        conn, path = cast(Tuple[_SCPConn, _SCPPath], path)
    elif isinstance(path, str) and sys.platform == 'win32' and \
            path[:1] in string.ascii_letters and \
            path[1:2] == ':': # pragma: no cover (win32)
        conn = None
    elif isinstance(path, str) and ':' in path:
        conn, path = path.split(':', 1)
    elif isinstance(path, bytes) and b':' in path:
        conn, path = path.split(b':', 1)
        conn = conn.decode('utf-8')
    elif isinstance(path, (bytes, str, PurePath)):
        conn = None
    else:
        conn = path
        path = b'.'

    if isinstance(conn, str):
        close_conn = True
        conn = await connect(conn, **kwargs)
    elif isinstance(conn, tuple):
        close_conn = True
        conn = await connect(*conn, **kwargs)
    else:
        close_conn = False

    return (cast(Optional['SSHClientConnection'], conn),
            cast(_SCPPath, path), close_conn)


async def _start_remote(conn: 'SSHClientConnection', source: bool,
                        must_be_dir: bool, preserve: bool,
                        recurse: bool, path: _SCPPath) -> \
        Tuple['SSHReader[bytes]', 'SSHWriter[bytes]']:
    """Start remote SCP server"""

    if isinstance(path, PurePath):
        path = str(path)

    if isinstance(path, str):
        path = path.encode('utf-8')

    command = (b'scp ' + (b'-f ' if source else b'-t ') +
               (b'-d ' if must_be_dir else b'') +
               (b'-p ' if preserve else b'') +
               (b'-r ' if recurse else b'') + path)

    conn.logger.get_child('sftp').info('Starting remote SCP, args: %s',
                                       command[4:])

    writer, reader, _ = await conn.open_session(command, encoding=None)

    return reader, writer


class _SCPArgs(argparse.Namespace):
    """SCP command line arguments"""

    path: str
    source: bool
    must_be_dir: bool
    preserve: bool
    recurse: bool


class _SCPArgParser(argparse.ArgumentParser):
    """A parser for SCP arguments"""

    def __init__(self) -> None:
        super().__init__(add_help=False)

        group = self.add_mutually_exclusive_group(required=True)
        group.add_argument('-f', dest='source', action='store_true')
        group.add_argument('-t', dest='source', action='store_false')

        self.add_argument('-d', dest='must_be_dir', action='store_true')
        self.add_argument('-p', dest='preserve', action='store_true')
        self.add_argument('-r', dest='recurse', action='store_true')
        self.add_argument('-v', dest='verbose', action='store_true')

        self.add_argument('path')

    def error(self, message: str) -> NoReturn:
        raise ValueError(message)

    def parse(self, command: str) -> _SCPArgs:
        """Parse an SCP command"""

        return self.parse_args(shlex.split(command)[1:], namespace=_SCPArgs())


class _SCPHandler:
    """SCP handler"""

    def __init__(self, reader: 'SSHReader[bytes]', writer: 'SSHWriter[bytes]',
                 error_handler: SFTPErrorHandler = None, server: bool = False):
        self._reader = reader
        self._writer = writer
        self._error_handler = error_handler
        self._server = server

        self._logger = reader.logger.get_child('sftp')

    async def __aenter__(self) -> Self: # pragma: no cover
        """Allow _SCPHandler to be used as an async context manager"""

        return self

    async def __aexit__(self, _exc_type: Optional[Type[BaseException]],
                        _exc_value: Optional[BaseException],
                        _traceback: Optional[TracebackType]) -> \
            bool: # pragma: no cover
        """Wait for file close when used as an async context manager"""

        await self.close()
        return False

    @property
    def logger(self) -> SSHLogger:
        """A logger associated with this SCP handler"""

        return self._logger

    async def await_response(self) -> Optional[Exception]:
        """Wait for an SCP response"""

        result = await self._reader.read(1)

        if result != b'\0':
            reason = await self._reader.readline()

            if not result or not reason.endswith(b'\n'):
                raise _scp_error(SFTPConnectionLost, 'Connection lost',
                                 fatal=True, suppress_send=True)

            if result not in b'\x01\x02':
                reason = result + reason

            return _scp_error(SFTPFailure, reason[:-1],
                              fatal=result != b'\x01', suppress_send=True)

        self.logger.debug1('Received SCP OK')

        return None

    def send_request(self, *args: bytes) -> None:
        """Send an SCP request"""

        request = b''.join(args)

        self.logger.debug1('Sending SCP request: %s', request)

        self._writer.write(request + b'\n')

    async def make_request(self, *args: bytes) -> None:
        """Send an SCP request and wait for a response"""

        self.send_request(*args)

        exc = await self.await_response()

        if exc:
            raise exc

    async def send_data(self, data: bytes) -> None:
        """Send SCP file data"""

        self.logger.debug1('Sending %s', plural(len(data), 'SCP data byte'))

        self._writer.write(data)
        await self._writer.drain()
        await asyncio.sleep(0)

    def send_ok(self) -> None:
        """Send an SCP OK response"""

        self.logger.debug1('Sending SCP OK')

        self._writer.write(b'\0')

    def send_error(self, exc: Exception) -> None:
        """Send an SCP error response"""

        if isinstance(exc, SFTPError):
            reason = exc.reason.encode('utf-8')
        elif isinstance(exc, OSError): # pragma: no branch (win32)
            reason = exc.strerror.encode('utf-8')

            filename = cast(BytesOrStr, exc.filename)

            if filename:
                if isinstance(filename, str): # pragma: no cover (win32)
                    filename = filename.encode('utf-8')

                reason += b': ' + filename
        else: # pragma: no cover (win32)
            reason = str(exc).encode('utf-8')

        fatal = cast(bool, getattr(exc, 'fatal', False))

        self.logger.debug1('Sending SCP %serror: %s',
                           'fatal ' if fatal else '', reason)

        self._writer.write((b'\x02' if fatal else b'\x01') +
                           b'scp: ' + reason + b'\n')

    async def recv_request(self) -> Tuple[Optional[bytes], Optional[bytes]]:
        """Receive SCP request"""

        request = await self._reader.readline()

        if not request:
            return None, None

        action, args = request[:1], request[1:-1]

        if action not in b'\x01\x02':
            self.logger.debug1('Received SCP request: %s%s', action, args)
        else:
            self.logger.debug1('Received SCP %serror: %s',
                               'fatal ' if action != b'\x01'  else '', args)

        return action, args


    async def recv_data(self, n: int) -> bytes:
        """Receive SCP file data"""

        data = await self._reader.read(n)

        self.logger.debug1('Received %s', plural(len(data), 'SCP data byte'))

        return data

    def handle_error(self, exc: Exception) -> None:
        """Handle an SCP error"""

        if isinstance(exc, BrokenPipeError):
            exc = _scp_error(SFTPConnectionLost, 'Connection lost',
                             fatal=True, suppress_send=True)

        if not getattr(exc, 'suppress_send', False):
            self.send_error(exc)

        self.logger.debug1('Handling SCP error: %s', str(exc))

        if self._error_handler and not getattr(exc, 'fatal', False):
            self._error_handler(exc)
        elif not self._server:
            raise exc

    async def close(self, cancelled: bool = False) -> None:
        """Close an SCP session"""

        self.logger.info('Stopping remote SCP')

        if cancelled:
            self._writer.channel.abort()
        else:
            if self._server:
                cast('SSHServerChannel', self._writer.channel).exit(0)
            else:
                self._writer.close()

            await self._writer.wait_closed()


class _SCPSource(_SCPHandler):
    """SCP handler for sending files"""

    def __init__(self, fs: _SCPFSProtocol, reader: 'SSHReader[bytes]',
                 writer: 'SSHWriter[bytes]', preserve: bool, recurse: bool,
                 block_size: int = _SCP_BLOCK_SIZE,
                 progress_handler: SFTPProgressHandler = None,
                 error_handler: SFTPErrorHandler = None, server: bool = False):
        super().__init__(reader, writer, error_handler, server)

        self._fs = fs
        self._preserve = preserve
        self._recurse = recurse
        self._block_size = block_size
        self._progress_handler = progress_handler

    async def _make_cd_request(self, action: bytes, attrs: SFTPAttrs,
                               size: int, path: bytes) -> None:
        """Make an SCP copy or dir request"""

        assert attrs.permissions is not None

        args = f'{attrs.permissions & 0o7777:04o} {size} '
        await self.make_request(action, args.encode('ascii'),
                                self._fs.basename(path))

    async def _make_t_request(self, attrs: SFTPAttrs) -> None:
        """Make an SCP time request"""

        self.logger.info('    Preserving attrs: %s',
                         SFTPAttrs(atime=attrs.atime, mtime=attrs.mtime))

        assert attrs.mtime is not None
        assert attrs.atime is not None

        args = f'{attrs.mtime} 0 {attrs.atime} 0'
        await self.make_request(b'T', args.encode('ascii'))

    async def _send_file(self, srcpath: bytes,
                         dstpath: bytes, attrs: SFTPAttrs) -> None:
        """Send a file over SCP"""

        assert attrs.size is not None

        file_obj = await self._fs.open(srcpath, 'rb')
        size = attrs.size
        local_exc = None
        offset = 0

        self.logger.info('  Sending file %s, size %d', srcpath, size)

        try:
            await self._make_cd_request(b'C', attrs, size, srcpath)

            if self._progress_handler and size == 0:
                self._progress_handler(srcpath, dstpath, 0, 0)

            while offset < size:
                blocklen = min(size - offset, self._block_size)

                if local_exc:
                    data = blocklen * b'\0'
                else:
                    try:
                        data = cast(bytes,
                                    await file_obj.read(blocklen, offset))

                        if not data:
                            raise _scp_error(SFTPFailure, 'Unexpected EOF')
                    except (OSError, SFTPError) as exc:
                        local_exc = exc

                await self.send_data(data)
                offset += len(data)

                if self._progress_handler:
                    self._progress_handler(srcpath, dstpath, offset, size)
        finally:
            await file_obj.close()

        if local_exc:
            self.send_error(local_exc)
            setattr(local_exc, 'suppress_send', True)
        else:
            self.send_ok()

        remote_exc = await self.await_response()
        final_exc = remote_exc or local_exc

        if final_exc:
            raise final_exc

    async def _send_dir(self, srcpath: bytes, dstpath: bytes,
                        attrs: SFTPAttrs) -> None:
        """Send directory over SCP"""

        self.logger.info('  Starting send of directory %s', srcpath)

        await self._make_cd_request(b'D', attrs, 0, srcpath)

        async for entry in self._fs.scandir(srcpath):
            name = cast(bytes, entry.filename)

            if name in (b'.', b'..'):
                continue

            await self._send_files(posixpath.join(srcpath, name),
                                   posixpath.join(dstpath, name),
                                   entry.attrs)

        await self.make_request(b'E')

        self.logger.info('  Finished send of directory %s', srcpath)

    async def _send_files(self, srcpath: bytes, dstpath: bytes,
                          attrs: SFTPAttrs) -> None:
        """Send files via SCP"""

        try:
            if self._preserve:
                await self._make_t_request(attrs)

            if self._recurse and attrs.type == FILEXFER_TYPE_DIRECTORY:
                await self._send_dir(srcpath, dstpath, attrs)
            elif attrs.type == FILEXFER_TYPE_REGULAR:
                await self._send_file(srcpath, dstpath, attrs)
            else:
                raise _scp_error(SFTPFailure, 'Not a regular file', srcpath)
        except (OSError, SFTPError, ValueError) as exc:
            self.handle_error(exc)

    async def run(self, srcpath: _SCPPath) -> None:
        """Start SCP transfer"""

        cancelled = False

        try:
            if isinstance(srcpath, PurePath):
                srcpath = str(srcpath)

            if isinstance(srcpath, str):
                srcpath = srcpath.encode('utf-8')

            exc = await self.await_response()

            if exc:
                raise exc

            for name in await SFTPGlob(self._fs).match(srcpath):
                await self._send_files(cast(bytes, name.filename),
                                            b'', name.attrs)
        except asyncio.CancelledError:
            cancelled = True
        except (OSError, SFTPError) as exc:
            self.handle_error(exc)
        finally:
            await self.close(cancelled)


class _SCPSink(_SCPHandler):
    """SCP handler for receiving files"""

    def __init__(self, fs: _SCPFSProtocol, reader: 'SSHReader[bytes]',
                 writer: 'SSHWriter[bytes]', must_be_dir: bool, preserve: bool,
                 recurse: bool, block_size: int = _SCP_BLOCK_SIZE,
                 progress_handler: SFTPProgressHandler = None,
                 error_handler: SFTPErrorHandler = None, server: bool = False):
        super().__init__(reader, writer, error_handler, server)

        self._fs = fs
        self._must_be_dir = must_be_dir
        self._preserve = preserve
        self._recurse = recurse
        self._block_size = block_size
        self._progress_handler = progress_handler

    async def _recv_file(self, srcpath: bytes,
                         dstpath: bytes, size: int) -> None:
        """Receive a file via SCP"""

        file_obj = await self._fs.open(dstpath, 'wb')
        local_exc = None
        offset = 0

        self.logger.info('  Receiving file %s, size %d', dstpath, size)

        try:
            self.send_ok()

            if self._progress_handler and size == 0:
                self._progress_handler(srcpath, dstpath, 0, 0)

            while offset < size:
                blocklen = min(size - offset, self._block_size)
                data = await self.recv_data(blocklen)

                if not data:
                    raise _scp_error(SFTPConnectionLost, 'Connection lost',
                                     fatal=True, suppress_send=True)

                if not local_exc:
                    try:
                        await file_obj.write(data, offset)
                    except (OSError, SFTPError) as exc:
                        local_exc = exc

                offset += len(data)

                if self._progress_handler:
                    self._progress_handler(srcpath, dstpath, offset, size)
        finally:
            await file_obj.close()

        remote_exc = await self.await_response()

        if local_exc:
            self.send_error(local_exc)
            setattr(local_exc, 'suppress_send',True)
        else:
            self.send_ok()

        final_exc = remote_exc or local_exc

        if final_exc:
            raise final_exc

    async def _recv_dir(self, srcpath: bytes, dstpath: bytes) -> None:
        """Receive a directory over SCP"""

        if not self._recurse:
            raise _scp_error(SFTPBadMessage,
                             'Directory received without recurse')

        self.logger.info('  Starting receive of directory %s', dstpath)

        if await self._fs.exists(dstpath):
            if not await self._fs.isdir(dstpath):
                raise _scp_error(SFTPFailure, 'Not a directory', dstpath)
        else:
            await self._fs.mkdir(dstpath)

        await self._recv_files(srcpath, dstpath)

        self.logger.info('  Finished receive of directory %s', dstpath)

    async def _recv_files(self, srcpath: bytes, dstpath: bytes) -> None:
        """Receive files over SCP"""

        self.send_ok()

        attrs = SFTPAttrs()

        while True:
            action, args = await self.recv_request()

            if not action:
                break

            assert args is not None

            try:
                if action in b'\x01\x02':
                    raise _scp_error(SFTPFailure, args,
                                     fatal=action != b'\x01',
                                     suppress_send=True)
                elif action == b'T':
                    if self._preserve:
                        attrs.atime, attrs.mtime = _parse_t_args(args)

                    self.send_ok()
                elif action == b'E':
                    self.send_ok()
                    break
                elif action in b'CD':
                    try:
                        attrs.permissions, size, name = _parse_cd_args(args)

                        new_srcpath = posixpath.join(srcpath, name)

                        if await self._fs.isdir(dstpath):
                            new_dstpath = posixpath.join(dstpath, name)
                        else:
                            new_dstpath = dstpath

                        if action == b'D':
                            await self._recv_dir(new_srcpath, new_dstpath)
                        else:
                            await self._recv_file(new_srcpath,
                                                  new_dstpath, size)

                        if self._preserve:
                            self.logger.info('    Preserving attrs: %s', attrs)
                            await self._fs.setstat(new_dstpath, attrs)
                    finally:
                        attrs = SFTPAttrs()
                else:
                    raise _scp_error(SFTPBadMessage, 'Unknown request')
            except (OSError, SFTPError) as exc:
                self.handle_error(exc)

    async def run(self, dstpath: _SCPPath) -> None:
        """Start SCP file receive"""

        cancelled = False

        try:
            if isinstance(dstpath, PurePath):
                dstpath = str(dstpath)

            if isinstance(dstpath, str):
                dstpath = dstpath.encode('utf-8')

            if self._must_be_dir and not await self._fs.isdir(dstpath):
                self.handle_error(_scp_error(SFTPFailure, 'Not a directory',
                                             dstpath))
            else:
                await self._recv_files(b'', dstpath)
        except asyncio.CancelledError:
            cancelled = True
        except (OSError, SFTPError, ValueError) as exc:
            self.handle_error(exc)
        finally:
            await self.close(cancelled)


class _SCPCopier:
    """SCP handler for remote-to-remote copies"""

    def __init__(self, src_reader: 'SSHReader[bytes]',
                 src_writer: 'SSHWriter[bytes]',
                 dst_reader: 'SSHReader[bytes]',
                 dst_writer: 'SSHWriter[bytes]',
                 block_size: int = _SCP_BLOCK_SIZE,
                 progress_handler: SFTPProgressHandler = None,
                 error_handler: SFTPErrorHandler = None):
        self._source = _SCPHandler(src_reader, src_writer)
        self._sink = _SCPHandler(dst_reader, dst_writer)
        self._logger = self._source.logger
        self._block_size = block_size
        self._progress_handler = progress_handler
        self._error_handler = error_handler

    @property
    def logger(self) -> SSHLogger:
        """A logger associated with this SCP handler"""

        return self._logger

    def _handle_error(self, exc: Exception) -> None:
        """Handle an SCP error"""

        if isinstance(exc, BrokenPipeError):
            exc = _scp_error(SFTPConnectionLost, 'Connection lost',
                             fatal=True, suppress_send=True)

        self.logger.debug1('Handling SCP error: %s', str(exc))

        if self._error_handler and not getattr(exc, 'fatal', False):
            self._error_handler(exc)
        else:
            raise exc

    async def _forward_response(self, src: _SCPHandler,
                                dst: _SCPHandler) -> Optional[Exception]:
        """Forward an SCP response between two remote SCP servers"""

        # pylint: disable=no-self-use

        try:
            exc = await src.await_response()

            if exc:
                dst.send_error(exc)
                return exc
            else:
                dst.send_ok()
                return None
        except OSError as exc:
            return exc

    async def _copy_file(self, path: bytes, size: int) -> None:
        """Copy a file from one remote SCP server to another"""

        self.logger.info('  Copying file %s, size %d', path, size)

        offset = 0

        if self._progress_handler and size == 0:
            self._progress_handler(path, path, 0, 0)

        while offset < size:
            blocklen = min(size - offset, self._block_size)
            data = await self._source.recv_data(blocklen)

            if not data:
                raise _scp_error(SFTPConnectionLost, 'Connection lost',
                                 fatal=True, suppress_send=True)

            await self._sink.send_data(data)
            offset += len(data)

            if self._progress_handler:
                self._progress_handler(path, path, offset, size)

        source_exc = await self._forward_response(self._source, self._sink)
        sink_exc = await self._forward_response(self._sink, self._source)

        exc = sink_exc or source_exc

        if exc:
            self._handle_error(exc)

    async def _copy_files(self) -> None:
        """Copy files from one SCP server to another"""

        exc = await self._forward_response(self._sink, self._source)

        if exc:
            self._handle_error(exc)

        pathlist: List[bytes] = []
        attrlist: List[SFTPAttrs] = []
        attrs = SFTPAttrs()

        while True:
            action, args = await self._source.recv_request()

            if not action:
                break

            assert args is not None

            self._sink.send_request(action, args)

            if action in b'\x01\x02':
                exc = _scp_error(SFTPFailure, args, fatal=action != b'\x01')
                self._handle_error(exc)
                continue

            exc = await self._forward_response(self._sink, self._source)

            if exc:
                self._handle_error(exc)
                continue

            if action in b'CD':
                try:
                    attrs.permissions, size, name = _parse_cd_args(args)

                    if action == b'C':
                        path = b'/'.join(pathlist + [name])
                        await self._copy_file(path, size)
                        self.logger.info('    Preserving attrs: %s', attrs)
                    else:
                        pathlist.append(name)
                        attrlist.append(attrs)
                        self.logger.info('  Starting copy of directory %s',
                                         b'/'.join(pathlist))
                finally:
                    attrs = SFTPAttrs()
            elif action == b'E':
                if pathlist:
                    self.logger.info('  Finished copy of directory %s',
                                     b'/'.join(pathlist))

                    pathlist.pop()
                    attrs = attrlist.pop()

                    self.logger.info('    Preserving attrs: %s', attrs)
                else:
                    break
            elif action == b'T':
                attrs.atime, attrs.mtime = _parse_t_args(args)
            else:
                raise _scp_error(SFTPBadMessage, 'Unknown SCP action')

    async def run(self) -> None:
        """Start SCP remote-to-remote transfer"""

        cancelled = False

        try:
            await self._copy_files()
        except asyncio.CancelledError:
            cancelled = True
        except (OSError, SFTPError) as exc:
            self._handle_error(exc)
        finally:
            await self._source.close(cancelled)
            await self._sink.close(cancelled)


async def scp(srcpaths: Union[_SCPConnPath, Sequence[_SCPConnPath]],
              dstpath: _SCPConnPath = None, *, preserve: bool = False,
              recurse: bool = False, block_size: int = _SCP_BLOCK_SIZE,
              progress_handler: SFTPProgressHandler = None,
              error_handler: SFTPErrorHandler = None, **kwargs) -> None:
    """Copy files using SCP

       This function is a coroutine which copies one or more files or
       directories using the SCP protocol. Source and destination paths
       can be `str` or `bytes` values to reference local files or can be
       a tuple of the form `(conn, path)` where `conn` is an open
       :class:`SSHClientConnection` to reference files and directories
       on a remote system.

       For convenience, a host name or tuple of the form `(host, port)`
       can be provided in place of the :class:`SSHClientConnection` to
       request that a new SSH connection be opened to a host using
       default connect arguments. A `str` or `bytes` value of the form
       `'host:path'` may also be used in place of the `(conn, path)`
       tuple to make a new connection to the requested host on the
       default SSH port.

       Either a single source path or a sequence of source paths can be
       provided, and each path can contain '*' and '?' wildcard characters
       which can be used to match multiple source files or directories.

       When copying a single file or directory, the destination path
       can be either the full path to copy data into or the path to an
       existing directory where the data should be placed. In the latter
       case, the base file name from the source path will be used as the
       destination name.

       When copying multiple files, the destination path must refer to
       a directory. If it doesn't already exist, a directory will be
       created with that name.

       If the destination path is an :class:`SSHClientConnection` without
       a path or the path provided is empty, files are copied into the
       default destination working directory.

       If preserve is `True`, the access and modification times and
       permissions of the original files and directories are set on the
       copied files. However, do to the timing of when this information
       is sent, the preserved access time will be what was set on the
       source file before the copy begins. So, the access time on the
       source file will no longer match the destination after the
       transfer completes.

       If recurse is `True` and the source path points at a directory,
       the entire subtree under that directory is copied.

       Symbolic links found on the source will have the contents of their
       target copied rather than creating a destination symbolic link.
       When using this option during a recursive copy, one needs to watch
       out for links that result in loops. SCP does not provide a
       mechanism for preserving links. If you need this, consider using
       SFTP instead.

       The block_size value controls the size of read and write operations
       issued to copy the files. It defaults to 256 KB.

       If progress_handler is specified, it will be called after each
       block of a file is successfully copied. The arguments passed to
       this handler will be the relative path of the file being copied,
       bytes copied so far, and total bytes in the file being copied. If
       multiple source paths are provided or recurse is set to `True`,
       the progress_handler will be called consecutively on each file
       being copied.

       If error_handler is specified and an error occurs during the copy,
       this handler will be called with the exception instead of it being
       raised. This is intended to primarily be used when multiple source
       paths are provided or when recurse is set to `True`, to allow
       error information to be collected without aborting the copy of the
       remaining files. The error handler can raise an exception if it
       wants the copy to completely stop. Otherwise, after an error, the
       copy will continue starting with the next file.

       If any other keyword arguments are specified, they will be passed
       to the AsyncSSH connect() call when attempting to open any new SSH
       connections needed to perform the file transfer.

       :param srcpaths:
           The paths of the source files or directories to copy
       :param dstpath: (optional)
           The path of the destination file or directory to copy into
       :param preserve: (optional)
           Whether or not to preserve the original file attributes
       :param recurse: (optional)
           Whether or not to recursively copy directories
       :param block_size: (optional)
           The block size to use for file reads and writes
       :param progress_handler: (optional)
           The function to call to report copy progress
       :param error_handler: (optional)
           The function to call when an error occurs
       :type preserve: `bool`
       :type recurse: `bool`
       :type block_size: `int`
       :type progress_handler: `callable`
       :type error_handler: `callable`

       :raises: | :exc:`OSError` if a local file I/O error occurs
                | :exc:`SFTPError` if the server returns an error
                | :exc:`ValueError` if both source and destination are local

    """

    if (isinstance(srcpaths, (bytes, str, PurePath)) or
            (isinstance(srcpaths, tuple) and len(srcpaths) == 2)):
        srcpaths = [srcpaths] # type: ignore

    srcpaths: Sequence[_SCPConnPath]

    must_be_dir = len(srcpaths) > 1

    dstconn, dstpath, close_dst = await _parse_path(dstpath, **kwargs)

    try:
        for srcpath in srcpaths:
            srcconn, srcpath, close_src = await _parse_path(srcpath, **kwargs)

            try:
                if srcconn and dstconn:
                    src_reader, src_writer = await _start_remote(
                        srcconn, True, must_be_dir, preserve, recurse, srcpath)

                    dst_reader, dst_writer = await _start_remote(
                        dstconn, False, must_be_dir, preserve, recurse, dstpath)

                    copier = _SCPCopier(src_reader, src_writer, dst_reader,
                                        dst_writer, block_size,
                                        progress_handler, error_handler)

                    await copier.run()
                elif srcconn:
                    reader, writer = await _start_remote(
                        srcconn, True, must_be_dir, preserve, recurse, srcpath)

                    sink = _SCPSink(local_fs, reader, writer, must_be_dir,
                                    preserve, recurse, block_size,
                                    progress_handler, error_handler)

                    await sink.run(dstpath)
                elif dstconn:
                    reader, writer = await _start_remote(
                        dstconn, False, must_be_dir, preserve, recurse, dstpath)

                    source = _SCPSource(local_fs, reader, writer,
                                        preserve, recurse, block_size,
                                        progress_handler, error_handler)

                    await source.run(srcpath)
                else:
                    raise ValueError('Local copy not supported')
            finally:
                if close_src:
                    assert srcconn is not None
                    srcconn.close()
                    await srcconn.wait_closed()
    finally:
        if close_dst:
            assert dstconn is not None
            dstconn.close()
            await dstconn.wait_closed()


async def _scp_handler(sftp_server: MaybeAwait[SFTPServer],
                       args: _SCPArgs, reader: 'SSHReader[bytes]',
                       writer: 'SSHWriter[bytes]') -> None:
    """Run an SCP server to handle this request"""

    if inspect.isawaitable(sftp_server):
        sftp_server = await sftp_server

    fs = SFTPServerFS(sftp_server)

    handler: Union[_SCPSource, _SCPSink]

    if args.source:
        handler = _SCPSource(fs, reader, writer, args.preserve,
                             args.recurse, error_handler=False, server=True)
    else:
        handler = _SCPSink(fs, reader, writer, args.must_be_dir,
                           args.preserve, args.recurse,
                           error_handler=False, server=True)

    try:
        await handler.run(args.path)
    finally:
        result = sftp_server.exit()

        if inspect.isawaitable(result):
            await result


def run_scp_server(sftp_server: MaybeAwait[SFTPServer], command: str,
                   stdin: 'SSHReader[bytes]', stdout: 'SSHWriter[bytes]',
                   stderr: 'SSHWriter[bytes]') -> MaybeAwait[None]:
    """Return a handler for an SCP server session"""

    try:
        args = _SCPArgParser().parse(command)
    except ValueError as exc:
        if inspect.iscoroutine(sftp_server):
            sftp_server.close()

        stdin.logger.info('Error starting SCP server: %s', str(exc))
        stderr.write(b'scp: ' + str(exc).encode('utf-8') + b'\n')
        cast('SSHServerChannel', stderr.channel).exit(1)
        return None

    stdin.logger.info('Starting SCP server, args: %s', command[4:].strip())

    return _scp_handler(sftp_server, args, stdin, stdout)



# Copyright (c) 2018-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""RSA key exchange handler"""

from hashlib import sha1, sha256
from typing import TYPE_CHECKING, Optional, cast

from .kex import Kex, register_kex_alg
from .misc import HashType, KeyExchangeFailed, ProtocolError
from .misc import get_symbol_names, randrange
from .packet import MPInt, String, SSHPacket
from .public_key import KeyImportError, SSHKey
from .public_key import decode_ssh_public_key, generate_private_key
from .rsa import RSAKey


if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .connection import SSHConnection, SSHClientConnection
    from .connection import SSHServerConnection


# SSH KEXRSA message values
MSG_KEXRSA_PUBKEY  = 30
MSG_KEXRSA_SECRET  = 31
MSG_KEXRSA_DONE    = 32


class _KexRSA(Kex):
    """Handler for RSA key exchange"""

    _handler_names = get_symbol_names(globals(), 'MSG_KEXRSA')

    def __init__(self, alg: bytes, conn: 'SSHConnection', hash_alg: HashType,
                 key_size: int, hash_size: int):
        super().__init__(alg, conn, hash_alg)

        self._key_size = key_size
        self._k_limit = 1 << (key_size - 2*hash_size - 49)

        self._host_key_data = b''

        self._trans_key: Optional[SSHKey] = None
        self._trans_key_data = b''

        self._k = 0
        self._encrypted_k = b''

    async def start(self) -> None:
        """Start RSA key exchange"""

        if self._conn.is_server():
            server_conn = cast('SSHServerConnection', self._conn)
            host_key = server_conn.get_server_host_key()
            assert host_key is not None
            self._host_key_data = host_key.public_data

            self._trans_key = generate_private_key(
                'ssh-rsa', key_size=self._key_size)
            self._trans_key_data = self._trans_key.public_data

            self.send_packet(MSG_KEXRSA_PUBKEY, String(self._host_key_data),
                             String(self._trans_key_data))

    def _compute_hash(self) -> bytes:
        """Compute a hash of key information associated with the connection"""

        hash_obj = self._hash_alg()
        hash_obj.update(self._conn.get_hash_prefix())
        hash_obj.update(String(self._host_key_data))
        hash_obj.update(String(self._trans_key_data))
        hash_obj.update(String(self._encrypted_k))
        hash_obj.update(MPInt(self._k))
        return hash_obj.digest()

    def _process_pubkey(self, _pkttype: int, _pktid: int,
                        packet: SSHPacket) -> None:
        """Process a KEXRSA pubkey message"""

        if self._conn.is_server():
            raise ProtocolError('Unexpected KEXRSA pubkey msg')

        self._host_key_data = packet.get_string()
        self._trans_key_data = packet.get_string()
        packet.check_end()

        try:
            pubkey = decode_ssh_public_key(self._trans_key_data)
        except KeyImportError:
            raise ProtocolError('Invalid KEXRSA pubkey msg') from None

        trans_key = cast(RSAKey, pubkey)
        self._k = randrange(self._k_limit)
        self._encrypted_k = \
            cast(bytes, trans_key.encrypt(MPInt(self._k), self.algorithm))

        self.send_packet(MSG_KEXRSA_SECRET, String(self._encrypted_k))

    def _process_secret(self, _pkttype: int, _pktid: int,
                        packet: SSHPacket) -> None:
        """Process a KEXRSA secret message"""

        if self._conn.is_client():
            raise ProtocolError('Unexpected KEXRSA secret msg')

        self._encrypted_k = packet.get_string()
        packet.check_end()

        trans_key = cast(RSAKey, self._trans_key)
        decrypted_k = trans_key.decrypt(self._encrypted_k, self.algorithm)
        if not decrypted_k:
            raise KeyExchangeFailed('Key exchange decryption failed')

        packet = SSHPacket(decrypted_k)
        self._k = packet.get_mpint()
        packet.check_end()

        server_conn = cast('SSHServerConnection', self._conn)
        host_key = server_conn.get_server_host_key()
        assert host_key is not None

        h = self._compute_hash()
        sig = host_key.sign(h)

        self.send_packet(MSG_KEXRSA_DONE, String(sig))

        self._conn.send_newkeys(MPInt(self._k), h)

    def _process_done(self, _pkttype: int, _pktid: int,
                      packet: SSHPacket) -> None:
        """Process a KEXRSA done message"""

        if self._conn.is_server():
            raise ProtocolError('Unexpected KEXRSA done msg')

        sig = packet.get_string()
        packet.check_end()

        client_conn = cast('SSHClientConnection', self._conn)
        host_key = client_conn.validate_server_host_key(self._host_key_data)

        h = self._compute_hash()
        if not host_key.verify(h, sig):
            raise KeyExchangeFailed('Key exchange hash mismatch')

        self._conn.send_newkeys(MPInt(self._k), h)

    _packet_handlers = {
        MSG_KEXRSA_PUBKEY: _process_pubkey,
        MSG_KEXRSA_SECRET: _process_secret,
        MSG_KEXRSA_DONE:   _process_done
    }


for _name, _hash_alg, _key_size, _hash_size, _default in (
        (b'rsa2048-sha256', sha256, 2048, 256, True),
        (b'rsa1024-sha1',   sha1,   1024, 160, False)):
    register_kex_alg(_name, _KexRSA, _hash_alg,
                     (_key_size, _hash_size), _default)



# Copyright (c) 2019-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""U2F ECDSA public key encryption handler"""

from hashlib import sha256
from typing import Optional, Tuple, cast

from .asn1 import der_encode, der_decode
from .crypto import ECDSAPublicKey
from .packet import Byte, MPInt, String, UInt32, SSHPacket
from .public_key import KeyExportError, SSHKey, SSHOpenSSHCertificateV01
from .public_key import register_public_key_alg, register_certificate_alg
from .public_key import register_sk_alg
from .sk import SSH_SK_ECDSA, SSH_SK_USER_PRESENCE_REQD
from .sk import sk_enroll, sk_sign, sk_webauthn_prefix, sk_use_webauthn


_PrivateKeyArgs = Tuple[bytes, bytes, str, int, bytes, bytes]
_PublicKeyArgs = Tuple[bytes, bytes, str]


class _SKECDSAKey(SSHKey):
    """Handler for U2F ECDSA public key encryption"""

    _key: ECDSAPublicKey

    use_executor = True

    def __init__(self, curve_id: bytes, public_value: bytes, application: str,
                 flags: int = 0, key_handle: Optional[bytes] = None,
                 reserved: bytes = b''):
        super().__init__(ECDSAPublicKey.construct(curve_id, public_value))

        self.algorithm = b'sk-ecdsa-sha2-' + curve_id + b'@openssh.com'
        self.sig_algorithms = (self.algorithm, b'webauthn-' + self.algorithm)
        self.all_sig_algorithms = set(self.sig_algorithms)

        self.use_webauthn = sk_use_webauthn

        self._application = application
        self._app_hash = sha256(application.encode('utf-8')).digest()
        self._flags = flags
        self._key_handle = key_handle
        self._reserved = reserved

    def __eq__(self, other: object) -> bool:
        # This isn't protected access - both objects are _SKECDSAKey instances
        # pylint: disable=protected-access

        return (isinstance(other, type(self)) and
                self._key.curve_id == other._key.curve_id and
                self._key.public_value == other._key.public_value and
                self._application == other._application and
                self._flags == other._flags and
                self._key_handle == other._key_handle and
                self._reserved == other._reserved)

    def __hash__(self) -> int:
        return hash((self._key.curve_id, self._key.public_value,
                     self._application, self._flags, self._key_handle,
                     self._reserved))

    @classmethod
    def generate(cls, algorithm: bytes, *, # type: ignore
                 application: str = 'ssh:', user: str = 'AsyncSSH',
                 pin: Optional[str] = None, resident: bool = False,
                 touch_required: bool = True) -> '_SKECDSAKey':
        """Generate a new SK ECDSA private key"""

        # pylint: disable=arguments-differ

        flags = SSH_SK_USER_PRESENCE_REQD if touch_required else 0

        public_value, key_handle = sk_enroll(SSH_SK_ECDSA, application,
                                             user, pin, resident)

        # Strip prefix and suffix of algorithm to get curve_id
        return cls(algorithm[14:-12], public_value, application,
                   flags, key_handle, b'')

    @classmethod
    def make_private(cls, key_params: object) -> SSHKey:
        """Construct a U2F ECDSA private key"""

        curve_id, public_value, application, flags, key_handle, reserved = \
            cast(_PrivateKeyArgs, key_params)

        return cls(curve_id, public_value, application,
                   flags, key_handle, reserved)

    @classmethod
    def make_public(cls, key_params: object) -> SSHKey:
        """Construct a U2F ECDSA public key"""

        curve_id, public_value, application = cast(_PublicKeyArgs, key_params)

        return cls(curve_id, public_value, application)

    @classmethod
    def decode_ssh_private(cls, packet: SSHPacket) -> _PrivateKeyArgs:
        """Decode an SSH format SK ECDSA private key"""

        curve_id = packet.get_string()
        public_value = packet.get_string()
        application = packet.get_string().decode('utf-8')
        flags = packet.get_byte()
        key_handle = packet.get_string()
        reserved = packet.get_string()

        return curve_id, public_value, application, flags, key_handle, reserved

    @classmethod
    def decode_ssh_public(cls, packet: SSHPacket) -> _PublicKeyArgs:
        """Decode an SSH format SK ECDSA public key"""

        curve_id = packet.get_string()
        public_value = packet.get_string()
        application = packet.get_string().decode('utf-8')

        return curve_id, public_value, application

    def encode_ssh_private(self) -> bytes:
        """Encode an SSH format SK ECDSA private key"""

        if self._key_handle is None:
            raise KeyExportError('Key is not private')

        return b''.join((String(self._key.curve_id),
                         String(self._key.public_value),
                         String(self._application), Byte(self._flags),
                         String(self._key_handle), String(self._reserved)))

    def encode_ssh_public(self) -> bytes:
        """Encode an SSH format SK ECDSA public key"""

        return b''.join((String(self._key.curve_id),
                         String(self._key.public_value),
                         String(self._application)))

    def encode_agent_cert_private(self) -> bytes:
        """Encode U2F ECDSA certificate private key data for agent"""

        if self._key_handle is None:
            raise KeyExportError('Key is not private')

        return b''.join((String(self._application), Byte(self._flags),
                         String(self._key_handle), String(self._reserved)))

    def sign_ssh(self, data: bytes, sig_algorithm: bytes) -> bytes:
        """Compute an SSH-encoded signature of the specified data"""

        if self._key_handle is None:
            raise ValueError('Key handle needed for signing')

        is_webauthn = sig_algorithm.startswith(b'webauthn')

        flags, counter, sig, client_data = sk_sign(data, self._application,
                                                   self._key_handle,
                                                   self._flags, is_webauthn)

        r, s = cast(Tuple[int, int], der_decode(sig))

        sig = String(MPInt(r) + MPInt(s)) + Byte(flags) + UInt32(counter)

        if is_webauthn:
            sig += String(self._application) + String(client_data) + String('')

        return sig

    def verify_ssh(self, data: bytes, sig_algorithm: bytes,
                   packet: SSHPacket) -> bool:
        """Verify an SSH-encoded signature of the specified data"""

        is_webauthn = sig_algorithm.startswith(b'webauthn')

        sig = packet.get_string()
        flags = packet.get_byte()
        counter = packet.get_uint32()

        if is_webauthn:
            _ = packet.get_string()            # origin
            client_data = packet.get_string()
            _ = packet.get_string()            # extensions

            prefix = sk_webauthn_prefix(data, self._application)

            if not client_data.startswith(prefix):
                return False

            data = client_data

        packet.check_end()

        if self._touch_required and not flags & SSH_SK_USER_PRESENCE_REQD:
            return False

        packet = SSHPacket(sig)
        r = packet.get_mpint()
        s = packet.get_mpint()
        packet.check_end()

        sig = der_encode((r, s))

        return self._key.verify(self._app_hash + Byte(flags) +
                                UInt32(counter) + sha256(data).digest(),
                                sig, 'sha256')


_algorithm = b'sk-ecdsa-sha2-nistp256@openssh.com'
_cert_algorithm = b'sk-ecdsa-sha2-nistp256-cert-v01@openssh.com'

register_sk_alg(SSH_SK_ECDSA, _SKECDSAKey, b'nistp256')

register_public_key_alg(_algorithm, _SKECDSAKey, True,
                        (_algorithm, b'webauthn-' + _algorithm))

register_certificate_alg(1, _algorithm, _cert_algorithm,
                         _SKECDSAKey, SSHOpenSSHCertificateV01, True)



# Copyright (c) 2013-2023 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""DSA public key encryption handler"""

from typing import Optional, Tuple, Union, cast

from .asn1 import ASN1DecodeError, ObjectIdentifier, der_encode, der_decode
from .crypto import DSAPrivateKey, DSAPublicKey
from .misc import all_ints
from .packet import MPInt, String, SSHPacket
from .public_key import SSHKey, SSHOpenSSHCertificateV01, KeyExportError
from .public_key import register_public_key_alg, register_certificate_alg
from .public_key import register_x509_certificate_alg


_PrivateKeyArgs = Tuple[int, int, int, int, int]
_PublicKeyArgs = Tuple[int, int, int, int]


class _DSAKey(SSHKey):
    """Handler for DSA public key encryption"""

    _key: Union[DSAPrivateKey, DSAPublicKey]

    algorithm = b'ssh-dss'
    default_x509_hash = 'sha256'
    pem_name = b'DSA'
    pkcs8_oid = ObjectIdentifier('1.2.840.10040.4.1')
    sig_algorithms = (algorithm,)
    x509_algorithms = (b'x509v3-' + algorithm,)
    all_sig_algorithms = set(sig_algorithms)

    def __eq__(self, other: object) -> bool:
        # This isn't protected access - both objects are _DSAKey instances
        # pylint: disable=protected-access

        return (isinstance(other, type(self)) and
                self._key.p == other._key.p and
                self._key.q == other._key.q and
                self._key.g == other._key.g and
                self._key.y == other._key.y and
                self._key.x == other._key.x)

    def __hash__(self) -> int:
        return hash((self._key.p, self._key.q, self._key.g,
                     self._key.y, self._key.x))

    @classmethod
    def generate(cls, algorithm: bytes) -> '_DSAKey': # type: ignore
        """Generate a new DSA private key"""

        # pylint: disable=arguments-differ,unused-argument

        return cls(DSAPrivateKey.generate(key_size=1024))

    @classmethod
    def make_private(cls, key_params: object) -> SSHKey:
        """Construct a DSA private key"""

        p, q, g, y, x = cast(_PrivateKeyArgs, key_params)

        return cls(DSAPrivateKey.construct(p, q, g, y, x))

    @classmethod
    def make_public(cls, key_params: object) -> SSHKey:
        """Construct a DSA public key"""

        p, q, g, y = cast(_PublicKeyArgs, key_params)

        return cls(DSAPublicKey.construct(p, q, g, y))

    @classmethod
    def decode_pkcs1_private(cls, key_data: object) -> \
            Optional[_PrivateKeyArgs]:
        """Decode a PKCS#1 format DSA private key"""

        if (isinstance(key_data, tuple) and len(key_data) == 6 and
                all_ints(key_data) and key_data[0] == 0):
            return cast(_PrivateKeyArgs, key_data[1:])
        else:
            return None

    @classmethod
    def decode_pkcs1_public(cls, key_data: object) -> \
            Optional[_PublicKeyArgs]:
        """Decode a PKCS#1 format DSA public key"""

        if (isinstance(key_data, tuple) and len(key_data) == 4 and
                all_ints(key_data)):
            y, p, q, g = key_data
            return p, q, g, y
        else:
            return None

    @classmethod
    def decode_pkcs8_private(cls, alg_params: object,
                             data: bytes) -> Optional[_PrivateKeyArgs]:
        """Decode a PKCS#8 format DSA private key"""

        try:
            x = der_decode(data)
        except ASN1DecodeError:
            return None

        if (isinstance(alg_params, tuple) and len(alg_params) == 3 and
                all_ints(alg_params) and isinstance(x, int)):
            p, q, g = alg_params
            y: int = pow(g, x, p)
            return p, q, g, y, x
        else:
            return None

    @classmethod
    def decode_pkcs8_public(cls, alg_params: object,
                            data: bytes) -> Optional[_PublicKeyArgs]:
        """Decode a PKCS#8 format DSA public key"""

        try:
            y = der_decode(data)
        except ASN1DecodeError:
            return None

        if (isinstance(alg_params, tuple) and len(alg_params) == 3 and
                all_ints(alg_params) and isinstance(y, int)):
            p, q, g = alg_params
            return p, q, g, y
        else:
            return None

    @classmethod
    def decode_ssh_private(cls, packet: SSHPacket) -> _PrivateKeyArgs:
        """Decode an SSH format DSA private key"""

        p = packet.get_mpint()
        q = packet.get_mpint()
        g = packet.get_mpint()
        y = packet.get_mpint()
        x = packet.get_mpint()

        return p, q, g, y, x

    @classmethod
    def decode_ssh_public(cls, packet: SSHPacket) -> _PublicKeyArgs:
        """Decode an SSH format DSA public key"""

        p = packet.get_mpint()
        q = packet.get_mpint()
        g = packet.get_mpint()
        y = packet.get_mpint()

        return p, q, g, y

    def encode_pkcs1_private(self) -> object:
        """Encode a PKCS#1 format DSA private key"""

        if not self._key.x:
            raise KeyExportError('Key is not private')

        return (0, self._key.p, self._key.q, self._key.g,
                self._key.y, self._key.x)

    def encode_pkcs1_public(self) -> object:
        """Encode a PKCS#1 format DSA public key"""

        return (self._key.y, self._key.p, self._key.q, self._key.g)

    def encode_pkcs8_private(self) -> Tuple[object, object]:
        """Encode a PKCS#8 format DSA private key"""

        if not self._key.x:
            raise KeyExportError('Key is not private')

        return (self._key.p, self._key.q, self._key.g), der_encode(self._key.x)

    def encode_pkcs8_public(self) -> Tuple[object, object]:
        """Encode a PKCS#8 format DSA public key"""

        return (self._key.p, self._key.q, self._key.g), der_encode(self._key.y)

    def encode_ssh_private(self) -> bytes:
        """Encode an SSH format DSA private key"""

        if not self._key.x:
            raise KeyExportError('Key is not private')

        return b''.join((MPInt(self._key.p), MPInt(self._key.q),
                         MPInt(self._key.g), MPInt(self._key.y),
                         MPInt(self._key.x)))

    def encode_ssh_public(self) -> bytes:
        """Encode an SSH format DSA public key"""

        return b''.join((MPInt(self._key.p), MPInt(self._key.q),
                         MPInt(self._key.g), MPInt(self._key.y)))

    def encode_agent_cert_private(self) -> bytes:
        """Encode DSA certificate private key data for agent"""

        if not self._key.x:
            raise KeyExportError('Key is not private')

        return MPInt(self._key.x)

    def sign_ssh(self, data: bytes, sig_algorithm: bytes) -> bytes:
        """Compute an SSH-encoded signature of the specified data"""

        # pylint: disable=unused-argument

        if not self._key.x:
            raise ValueError('Private key needed for signing')

        sig = der_decode(self._key.sign(data, 'sha1'))
        r, s = cast(Tuple[int, int], sig)
        return String(r.to_bytes(20, 'big') + s.to_bytes(20, 'big'))

    def verify_ssh(self, data: bytes, sig_algorithm: bytes,
                   packet: SSHPacket) -> bool:
        """Verify an SSH-encoded signature of the specified data"""

        # pylint: disable=unused-argument

        sig = packet.get_string()
        packet.check_end()

        if len(sig) != 40:
            return False

        r = int.from_bytes(sig[:20], 'big')
        s = int.from_bytes(sig[20:], 'big')

        return self._key.verify(data, der_encode((r, s)), 'sha1')


register_public_key_alg(b'ssh-dss', _DSAKey, False)

register_certificate_alg(1, b'ssh-dss', b'ssh-dss-cert-v01@openssh.com',
                         _DSAKey, SSHOpenSSHCertificateV01, False)

for alg in _DSAKey.x509_algorithms:
    register_x509_certificate_alg(alg, False)



# Copyright (c) 2017-2023 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""Miscellaneous PyCA utility classes and functions"""

from typing import Callable, Mapping, Union

from cryptography.hazmat.primitives.asymmetric import dsa, ec, rsa
from cryptography.hazmat.primitives.asymmetric import ed25519, ed448
from cryptography.hazmat.primitives.hashes import HashAlgorithm
from cryptography.hazmat.primitives.hashes import MD5, SHA1, SHA224
from cryptography.hazmat.primitives.hashes import SHA256, SHA384, SHA512


PyCAPrivateKey = Union[dsa.DSAPrivateKey, rsa.RSAPrivateKey,
                       ec.EllipticCurvePrivateKey,
                       ed25519.Ed25519PrivateKey, ed448.Ed448PrivateKey]

PyCAPublicKey = Union[dsa.DSAPublicKey, rsa.RSAPublicKey,
                      ec.EllipticCurvePublicKey,
                      ed25519.Ed25519PublicKey, ed448.Ed448PublicKey]

PyCAKey = Union[PyCAPrivateKey, PyCAPublicKey]


hashes: Mapping[str, Callable[[], HashAlgorithm]] = {
    str(h.name): h for h in (MD5, SHA1, SHA224, SHA256, SHA384, SHA512)
}


class CryptoKey:
    """Base class for PyCA private/public keys"""

    def __init__(self, pyca_key: PyCAKey):
        self._pyca_key = pyca_key

    @property
    def pyca_key(self) -> PyCAKey:
        """Return the PyCA object associated with this key"""

        return self._pyca_key

    def sign(self, data: bytes, hash_name: str = '') -> bytes:
        """Sign a block of data"""

        # pylint: disable=no-self-use
        raise RuntimeError # pragma: no cover

    def verify(self, data: bytes, sig: bytes, hash_name: str = '') -> bool:
        """Verify the signature on a block of data"""

        # pylint: disable=no-self-use
        raise RuntimeError # pragma: no cover



# Copyright (c) 2013-2021 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""SSH constants"""

# Default language for error messages
DEFAULT_LANG                        = 'en-US'

# Default SSH listening port
DEFAULT_PORT                        = 22

# SSH message codes
MSG_DISCONNECT                      = 1
MSG_IGNORE                          = 2
MSG_UNIMPLEMENTED                   = 3
MSG_DEBUG                           = 4
MSG_SERVICE_REQUEST                 = 5
MSG_SERVICE_ACCEPT                  = 6
MSG_EXT_INFO                        = 7

MSG_KEXINIT                         = 20
MSG_NEWKEYS                         = 21

MSG_KEX_FIRST                       = 30
MSG_KEX_LAST                        = 49

MSG_USERAUTH_REQUEST                = 50
MSG_USERAUTH_FAILURE                = 51
MSG_USERAUTH_SUCCESS                = 52
MSG_USERAUTH_BANNER                 = 53

MSG_USERAUTH_FIRST                  = 60
MSG_USERAUTH_LAST                   = 79

MSG_GLOBAL_REQUEST                  = 80
MSG_REQUEST_SUCCESS                 = 81
MSG_REQUEST_FAILURE                 = 82

MSG_CHANNEL_OPEN                    = 90
MSG_CHANNEL_OPEN_CONFIRMATION       = 91
MSG_CHANNEL_OPEN_FAILURE            = 92

MSG_CHANNEL_WINDOW_ADJUST           = 93
MSG_CHANNEL_DATA                    = 94
MSG_CHANNEL_EXTENDED_DATA           = 95
MSG_CHANNEL_EOF                     = 96
MSG_CHANNEL_CLOSE                   = 97
MSG_CHANNEL_REQUEST                 = 98
MSG_CHANNEL_SUCCESS                 = 99
MSG_CHANNEL_FAILURE                 = 100

# Messages 90-92 are excluded here as they relate to opening a new channel
MSG_CHANNEL_FIRST                   = 93
MSG_CHANNEL_LAST                    = 127

# SSH disconnect reason codes
DISC_HOST_NOT_ALLOWED_TO_CONNECT    = 1
DISC_PROTOCOL_ERROR                 = 2
DISC_KEY_EXCHANGE_FAILED            = 3
DISC_RESERVED                       = 4
DISC_MAC_ERROR                      = 5
DISC_COMPRESSION_ERROR              = 6
DISC_SERVICE_NOT_AVAILABLE          = 7
DISC_PROTOCOL_VERSION_NOT_SUPPORTED = 8
DISC_HOST_KEY_NOT_VERIFIABLE        = 9
DISC_CONNECTION_LOST                = 10
DISC_BY_APPLICATION                 = 11
DISC_TOO_MANY_CONNECTIONS           = 12
DISC_AUTH_CANCELLED_BY_USER         = 13
DISC_NO_MORE_AUTH_METHODS_AVAILABLE = 14
DISC_ILLEGAL_USER_NAME              = 15

DISC_HOST_KEY_NOT_VERIFYABLE        = 9   # Error in naming, left here to not
                                          # break backward compatibility

# SSH channel open failure reason codes
OPEN_ADMINISTRATIVELY_PROHIBITED    = 1
OPEN_CONNECT_FAILED                 = 2
OPEN_UNKNOWN_CHANNEL_TYPE           = 3
OPEN_RESOURCE_SHORTAGE              = 4

# Internal failure reason codes
OPEN_REQUEST_X11_FORWARDING_FAILED  = 0xfffffffd
OPEN_REQUEST_PTY_FAILED             = 0xfffffffe
OPEN_REQUEST_SESSION_FAILED         = 0xffffffff

# SFTPv3-v5 packet types
FXP_INIT                            = 1
FXP_VERSION                         = 2
FXP_OPEN                            = 3
FXP_CLOSE                           = 4
FXP_READ                            = 5
FXP_WRITE                           = 6
FXP_LSTAT                           = 7
FXP_FSTAT                           = 8
FXP_SETSTAT                         = 9
FXP_FSETSTAT                        = 10
FXP_OPENDIR                         = 11
FXP_READDIR                         = 12
FXP_REMOVE                          = 13
FXP_MKDIR                           = 14
FXP_RMDIR                           = 15
FXP_REALPATH                        = 16
FXP_STAT                            = 17
FXP_RENAME                          = 18
FXP_READLINK                        = 19
FXP_SYMLINK                         = 20
FXP_STATUS                          = 101
FXP_HANDLE                          = 102
FXP_DATA                            = 103
FXP_NAME                            = 104
FXP_ATTRS                           = 105
FXP_EXTENDED                        = 200
FXP_EXTENDED_REPLY                  = 201

# SFTPv6 packet types
FXP_LINK                            = 21
FXP_BLOCK                           = 22
FXP_UNBLOCK                         = 23

# SFTPv3 open flags
FXF_READ                            = 0x00000001
FXF_WRITE                           = 0x00000002
FXF_APPEND                          = 0x00000004
FXF_CREAT                           = 0x00000008
FXF_TRUNC                           = 0x00000010
FXF_EXCL                            = 0x00000020

# SFTPv4 open flags
FXF_TEXT                            = 0x00000040

# SFTPv5 open flags
FXF_ACCESS_DISPOSITION              = 0x00000007
FXF_CREATE_NEW                      = 0x00000000
FXF_CREATE_TRUNCATE                 = 0x00000001
FXF_OPEN_EXISTING                   = 0x00000002
FXF_OPEN_OR_CREATE                  = 0x00000003
FXF_TRUNCATE_EXISTING               = 0x00000004
FXF_APPEND_DATA                     = 0x00000008
FXF_APPEND_DATA_ATOMIC              = 0x00000010
FXF_TEXT_MODE                       = 0x00000020
FXF_BLOCK_READ                      = 0x00000040
FXF_BLOCK_WRITE                     = 0x00000080
FXF_BLOCK_DELETE                    = 0x00000100

# SFTPv6 open flags
FXF_BLOCK_ADVISORY                  = 0x00000200
FXF_NOFOLLOW                        = 0x00000400
FXF_DELETE_ON_CLOSE                 = 0x00000800
FXF_ACCESS_AUDIT_ALARM_INFO         = 0x00001000
FXF_ACCESS_BACKUP                   = 0x00002000
FXF_BACKUP_STREAM                   = 0x00004000
FXF_OVERRIDE_OWNER                  = 0x00008000

# SFTPv5-v6 ACE mask values used in desired-access
ACE4_READ_DATA                      = 0x00000001
ACE4_WRITE_DATA                     = 0x00000002
ACE4_APPEND_DATA                    = 0x00000004
ACE4_READ_ATTRIBUTES                = 0x00000080
ACE4_WRITE_ATTRIBUTES               = 0x00000100

# SFTPv3 attribute flags
FILEXFER_ATTR_SIZE                  = 0x00000001
FILEXFER_ATTR_UIDGID                = 0x00000002
FILEXFER_ATTR_PERMISSIONS           = 0x00000004
FILEXFER_ATTR_ACMODTIME             = 0x00000008
FILEXFER_ATTR_EXTENDED              = 0x80000000
FILEXFER_ATTR_DEFINED_V3            = 0x8000000f

# SFTPv4 attribute flags
FILEXFER_ATTR_ACCESSTIME            = 0x00000008
FILEXFER_ATTR_CREATETIME            = 0x00000010
FILEXFER_ATTR_MODIFYTIME            = 0x00000020
FILEXFER_ATTR_ACL                   = 0x00000040
FILEXFER_ATTR_OWNERGROUP            = 0x00000080
FILEXFER_ATTR_SUBSECOND_TIMES       = 0x00000100
FILEXFER_ATTR_DEFINED_V4            = 0x800001fd

# SFTPv5 attribute flags
FILEXFER_ATTR_BITS                  = 0x00000200
FILEXFER_ATTR_DEFINED_V5            = 0x800003fd

# SFTPv6 attribute flags
FILEXFER_ATTR_ALLOCATION_SIZE       = 0x00000400
FILEXFER_ATTR_TEXT_HINT             = 0x00000800
FILEXFER_ATTR_MIME_TYPE             = 0x00001000
FILEXFER_ATTR_LINK_COUNT            = 0x00002000
FILEXFER_ATTR_UNTRANSLATED_NAME     = 0x00004000
FILEXFER_ATTR_CTIME                 = 0x00008000
FILEXFER_ATTR_DEFINED_V6            = 0x8000fffd

# SFTPv4 file types
FILEXFER_TYPE_REGULAR               = 1
FILEXFER_TYPE_DIRECTORY             = 2
FILEXFER_TYPE_SYMLINK               = 3
FILEXFER_TYPE_SPECIAL               = 4
FILEXFER_TYPE_UNKNOWN               = 5

# SFTPv5 file types
FILEXFER_TYPE_SOCKET                = 6
FILEXFER_TYPE_CHAR_DEVICE           = 7
FILEXFER_TYPE_BLOCK_DEVICE          = 8
FILEXFER_TYPE_FIFO                  = 9

# SFTPv5 attrib bits
FILEXFER_ATTR_BITS_READONLY         = 0x00000001
FILEXFER_ATTR_BITS_SYSTEM           = 0x00000002
FILEXFER_ATTR_BITS_HIDDEN           = 0x00000004
FILEXFER_ATTR_BITS_CASE_INSENSITIVE = 0x00000008
FILEXFER_ATTR_BITS_ARCHIVE          = 0x00000010
FILEXFER_ATTR_BITS_ENCRYPTED        = 0x00000020
FILEXFER_ATTR_BITS_COMPRESSED       = 0x00000040
FILEXFER_ATTR_BITS_SPARSE           = 0x00000080
FILEXFER_ATTR_BITS_APPEND_ONLY      = 0x00000100
FILEXFER_ATTR_BITS_IMMUTABLE        = 0x00000200
FILEXFER_ATTR_BITS_SYNC             = 0x00000400

# SFTPv6 attrib bits
FILEXFER_ATTR_BITS_TRANSLATION_ERR  = 0x00000800

# SFTPv6 text hint flags
FILEXFER_ATTR_KNOWN_TEXT            = 0
FILEXFER_ATTR_GUESSED_TEXT          = 1
FILEXFER_ATTR_KNOWN_BINARY          = 2
FILEXFER_ATTR_GUESSED_BINARY        = 3

# SFTPv5 rename flags
FXR_OVERWRITE                       = 0x00000001
FXR_ATOMIC                          = 0x00000002
FXR_NATIVE                          = 0x00000004

# SFTPv6 realpath control byte
FXRP_NO_CHECK                       = 1
FXRP_STAT_IF_EXISTS                 = 2
FXRP_STAT_ALWAYS                    = 3

# OpenSSH statvfs attribute flags
FXE_STATVFS_ST_RDONLY               = 0x1
FXE_STATVFS_ST_NOSUID               = 0x2

# SFTPv3 error codes
FX_OK                               = 0
FX_EOF                              = 1
FX_NO_SUCH_FILE                     = 2
FX_PERMISSION_DENIED                = 3
FX_FAILURE                          = 4
FX_BAD_MESSAGE                      = 5
FX_NO_CONNECTION                    = 6
FX_CONNECTION_LOST                  = 7
FX_OP_UNSUPPORTED                   = 8
FX_V3_END                           = FX_OP_UNSUPPORTED

# SFTPv4 error codes
FX_INVALID_HANDLE                   = 9
FX_NO_SUCH_PATH                     = 10
FX_FILE_ALREADY_EXISTS              = 11
FX_WRITE_PROTECT                    = 12
FX_NO_MEDIA                         = 13
FX_V4_END                           = FX_NO_MEDIA

# SFTPv5 error codes
FX_NO_SPACE_ON_FILESYSTEM           = 14
FX_QUOTA_EXCEEDED                   = 15
FX_UNKNOWN_PRINCIPAL                = 16
FX_LOCK_CONFLICT                    = 17
FX_V5_END                           = FX_LOCK_CONFLICT

# SFTPv6 error codes
FX_DIR_NOT_EMPTY                    = 18
FX_NOT_A_DIRECTORY                  = 19
FX_INVALID_FILENAME                 = 20
FX_LINK_LOOP                        = 21
FX_CANNOT_DELETE                    = 22
FX_INVALID_PARAMETER                = 23
FX_FILE_IS_A_DIRECTORY              = 24
FX_BYTE_RANGE_LOCK_CONFLICT         = 25
FX_BYTE_RANGE_LOCK_REFUSED          = 26
FX_DELETE_PENDING                   = 27
FX_FILE_CORRUPT                     = 28
FX_OWNER_INVALID                    = 29
FX_GROUP_INVALID                    = 30
FX_NO_MATCHING_BYTE_RANGE_LOCK      = 31
FX_V6_END                           = FX_NO_MATCHING_BYTE_RANGE_LOCK

# SSH channel data type codes
EXTENDED_DATA_STDERR                = 1

# SSH pty mode opcodes
PTY_OP_END                          = 0
PTY_VINTR                           = 1
PTY_VQUIT                           = 2
PTY_VERASE                          = 3
PTY_VKILL                           = 4
PTY_VEOF                            = 5
PTY_VEOL                            = 6
PTY_VEOL2                           = 7
PTY_VSTART                          = 8
PTY_VSTOP                           = 9
PTY_VSUSP                           = 10
PTY_VDSUSP                          = 11
PTY_VREPRINT                        = 12
PTY_WERASE                          = 13
PTY_VLNEXT                          = 14
PTY_VFLUSH                          = 15
PTY_VSWTCH                          = 16
PTY_VSTATUS                         = 17
PTY_VDISCARD                        = 18
PTY_IGNPAR                          = 30
PTY_PARMRK                          = 31
PTY_INPCK                           = 32
PTY_ISTRIP                          = 33
PTY_INLCR                           = 34
PTY_IGNCR                           = 35
PTY_ICRNL                           = 36
PTY_IUCLC                           = 37
PTY_IXON                            = 38
PTY_IXANY                           = 39
PTY_IXOFF                           = 40
PTY_IMAXBEL                         = 41
PTY_IUTF8                           = 42
PTY_ISIG                            = 50
PTY_ICANON                          = 51
PTY_XCASE                           = 52
PTY_ECHO                            = 53
PTY_ECHOE                           = 54
PTY_ECHOK                           = 55
PTY_ECHONL                          = 56
PTY_NOFLSH                          = 57
PTY_TOSTOP                          = 58
PTY_IEXTEN                          = 59
PTY_ECHOCTL                         = 60
PTY_ECHOKE                          = 61
PTY_PENDIN                          = 62
PTY_OPOST                           = 70
PTY_OLCUC                           = 71
PTY_ONLCR                           = 72
PTY_OCRNL                           = 73
PTY_ONOCR                           = 74
PTY_ONLRET                          = 75
PTY_CS7                             = 90
PTY_CS8                             = 91
PTY_PARENB                          = 92
PTY_PARODD                          = 93
PTY_OP_ISPEED                       = 128
PTY_OP_OSPEED                       = 129
PTY_OP_RESERVED                     = 160



# Copyright (c) 2018-2023 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""SOCKS forwarding support"""

from ipaddress import ip_address
from typing import TYPE_CHECKING, Callable, Optional

from .forward import SSHForwarderCoro, SSHLocalForwarder
from .session import DataType


if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .connection import SSHConnection


_RecvHandler = Optional[Callable[[bytes], None]]


SOCKS4                  = 0x04
SOCKS5                  = 0x05

SOCKS_CONNECT           = 0x01

SOCKS4_OK               = 0x5a
SOCKS5_OK               = 0x00

SOCKS5_AUTH_NONE        = 0x00

SOCKS5_ADDR_IPV4        = 0x01
SOCKS5_ADDR_HOSTNAME    = 0x03
SOCKS5_ADDR_IPV6        = 0x04

SOCKS4_OK_RESPONSE      = bytes((0, SOCKS4_OK, 0, 0, 0, 0, 0, 0))
SOCKS5_OK_RESPONSE_HDR  = bytes((SOCKS5, SOCKS5_OK, 0))

_socks5_addr_len = { SOCKS5_ADDR_IPV4: 4, SOCKS5_ADDR_IPV6: 16 }


class SSHSOCKSForwarder(SSHLocalForwarder):
    """SOCKS dynamic port forwarding connection handler"""

    def __init__(self, conn: 'SSHConnection', coro: SSHForwarderCoro):
        super().__init__(conn, coro)

        self._inpbuf = b''
        self._bytes_needed = 2
        self._recv_handler: _RecvHandler = self._recv_version
        self._addrtype = 0
        self._host = ''
        self._port = 0

    def _connect(self) -> None:
        """Send request to open a new tunnel connection"""

        assert self._transport is not None

        self._recv_handler = None

        orig_host, orig_port = self._transport.get_extra_info('peername')[:2]
        self.forward(self._host, self._port, orig_host, orig_port)

    def _send_socks4_ok(self) -> None:
        """Send SOCKS4 success response"""

        assert self._transport is not None

        self._transport.write(SOCKS4_OK_RESPONSE)

    def _send_socks5_ok(self) -> None:
        """Send SOCKS5 success response"""

        assert self._transport is not None

        addrlen = _socks5_addr_len[self._addrtype] + 2

        self._transport.write(SOCKS5_OK_RESPONSE_HDR +
                              bytes((self._addrtype,)) +
                              addrlen * b'\0')

    def _recv_version(self, data: bytes) -> None:
        """Parse SOCKS version"""

        if data[0] == SOCKS4:
            if data[1] == SOCKS_CONNECT:
                self._bytes_needed = 6
                self._recv_handler = self._recv_socks4_addr
            else:
                self.close()
        elif data[0] == SOCKS5:
            self._bytes_needed = data[1]
            self._recv_handler = self._recv_socks5_authlist
        else:
            self.close()

    def _recv_socks4_addr(self, data: bytes) -> None:
        """Parse SOCKSv4 address and port"""

        self._port = (data[0] << 8) + data[1]

        # If address is 0.0.0.x, read a hostname later
        if data[2:5] != b'\0\0\0' or data[5] == 0:
            self._host = str(ip_address(data[2:]))

        self._bytes_needed = -1
        self._recv_handler = self._recv_socks4_user

    def _recv_socks4_user(self, data: bytes) -> None:
        """Parse SOCKSv4 username"""

        # pylint: disable=unused-argument

        if self._host:
            self._send_socks4_ok()
            self._connect()
        else:
            self._bytes_needed = -1
            self._recv_handler = self._recv_socks4_hostname

    def _recv_socks4_hostname(self, data: bytes) -> None:
        """Parse SOCKSv4 hostname"""

        try:
            self._host = data.decode('utf-8')
        except UnicodeDecodeError:
            self.close()
            return

        self._send_socks4_ok()
        self._connect()

    def _recv_socks5_authlist(self, data: bytes) -> None:
        """Parse SOCKSv5 list of authentication methods"""

        assert self._transport is not None

        if SOCKS5_AUTH_NONE in data:
            self._transport.write(bytes((SOCKS5, SOCKS5_AUTH_NONE)))

            self._bytes_needed = 4
            self._recv_handler = self._recv_socks5_command
        else:
            self.close()

    def _recv_socks5_command(self, data: bytes) -> None:
        """Parse SOCKSv5 command"""

        if data[0] == SOCKS5 and data[1] == SOCKS_CONNECT and data[2] == 0:
            if data[3] == SOCKS5_ADDR_HOSTNAME:
                self._bytes_needed = 1
                self._recv_handler = self._recv_socks5_hostlen
                self._addrtype = SOCKS5_ADDR_IPV4
            else:
                addrlen = _socks5_addr_len.get(data[3])

                if addrlen:
                    self._bytes_needed = addrlen
                    self._recv_handler = self._recv_socks5_addr
                    self._addrtype = data[3]
                else:
                    self.close()
        else:
            self.close()

    def _recv_socks5_addr(self, data: bytes) -> None:
        """Parse SOCKSv5 address"""

        self._host = str(ip_address(data))

        self._bytes_needed = 2
        self._recv_handler = self._recv_socks5_port

    def _recv_socks5_hostlen(self, data: bytes) -> None:
        """Parse SOCKSv5 host length"""

        self._bytes_needed = data[0]
        self._recv_handler = self._recv_socks5_host

    def _recv_socks5_host(self, data: bytes) -> None:
        """Parse SOCKSv5 host"""

        try:
            self._host = data.decode('utf-8')
        except UnicodeDecodeError:
            self.close()
            return

        self._bytes_needed = 2
        self._recv_handler = self._recv_socks5_port

    def _recv_socks5_port(self, data: bytes) -> None:
        """Parse SOCKSv5 port"""

        self._port = (data[0] << 8) + data[1]
        self._send_socks5_ok()
        self._connect()

    def data_received(self, data: bytes, datatype: DataType = None) -> None:
        """Handle incoming data from the SOCKS client"""

        if self._recv_handler:
            self._inpbuf += data

            while self._recv_handler: # type: ignore[truthy-function]
                if self._bytes_needed < 0:
                    idx = self._inpbuf.find(b'\0')
                    if idx >= 0:
                        data = self._inpbuf[:idx]
                        self._inpbuf = self._inpbuf[idx+1:]
                        self._recv_handler(data)
                    elif len(self._inpbuf) > 255:
                        # SOCKSv4 user or hostname too long
                        self.close()
                        return
                    else:
                        return
                else:
                    if len(self._inpbuf) >= self._bytes_needed:
                        data = self._inpbuf[:self._bytes_needed]
                        self._inpbuf = self._inpbuf[self._bytes_needed:]
                        self._recv_handler(data)
                    else:
                        return

            data = self._inpbuf
            self._inpbuf = b''

        if data:
            super().data_received(data, datatype)



# Copyright (c) 2020-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""Parser for OpenSSH config files"""

import os
import re
import shlex
import socket
import subprocess

from hashlib import sha1
from pathlib import Path, PurePath
from subprocess import DEVNULL
from typing import Callable, Dict, List, NoReturn, Optional, Sequence
from typing import Set, Tuple, Union, cast

from .constants import DEFAULT_PORT
from .logging import logger
from .misc import DefTuple, FilePath, ip_address
from .pattern import HostPatternList, WildcardPatternList


ConfigPaths = Union[None, FilePath, Sequence[FilePath]]


_token_pattern = re.compile(r'%(.)')
_env_pattern = re.compile(r'\${(.*)}')


def _exec(cmd: str) -> bool:
    """Execute a command and return if exit status is 0"""

    return subprocess.run(cmd, check=False, shell=True, stdin=DEVNULL,
                          stdout=DEVNULL, stderr=DEVNULL).returncode == 0


class ConfigParseError(ValueError):
    """Configuration parsing exception"""


class SSHConfig:
    """Settings from an OpenSSH config file"""

    _conditionals = {'match'}
    _no_split: Set[str] = set()
    _percent_expand = {'AuthorizedKeysFile'}
    _handlers: Dict[str, Tuple[str, Callable]] = {}

    def __init__(self, last_config: Optional['SSHConfig'], reload: bool,
                 canonical: bool, final: bool):
        if last_config:
            self._last_options = last_config.get_options(reload)
        else:
            self._last_options = {}

        self._canonical = canonical
        self._final = True if final else None
        self._default_path = Path('~', '.ssh').expanduser()
        self._path = Path()
        self._line_no = 0
        self._matching = True
        self._options = self._last_options.copy()
        self._tokens: Dict[str, str] = {}

        self.loaded = False

    def _error(self, reason: str) -> NoReturn:
        """Raise a configuration parsing error"""

        raise ConfigParseError(f'{self._path} line {self._line_no}: {reason}')

    def _match_val(self, match: str) -> object:
        """Return the value to match against in a match condition"""

        raise NotImplementedError

    def _set_tokens(self) -> None:
        """Set the tokens available for percent expansion"""

        raise NotImplementedError

    def _expand_token(self, match):
        """Expand a percent token reference"""

        try:
            token = match.group(1)
            return self._tokens[token]
        except KeyError:
            if token == 'd':
                raise ConfigParseError('Home directory is '
                                       'not available') from None
            elif token == 'i':
                raise ConfigParseError('User id not available') from None
            else:
                raise ConfigParseError('Invalid token expansion: ' +
                                       token) from None

    @staticmethod
    def _expand_env(match):
        """Expand an environment variable reference"""

        try:
            var = match.group(1)
            return os.environ[var]
        except KeyError:
            raise ConfigParseError('Invalid environment expansion: ' +
                                   var) from None

    def _expand_val(self, value: str) -> str:
        """Perform percent token and environment expansion on a string"""

        return _env_pattern.sub(self._expand_env,
                                _token_pattern.sub(self._expand_token, value))

    def _include(self, option: str, args: List[str]) -> None:
        """Read config from a list of other config files"""

        # pylint: disable=unused-argument

        old_path = self._path

        for pattern in args:
            path = Path(pattern).expanduser()

            if path.anchor:
                pattern = str(Path(*path.parts[1:]))
                path = Path(path.anchor)
            else:
                path = self._default_path

            paths = list(path.glob(pattern))

            if not paths:
                logger.debug1(f'Config pattern "{pattern}" matched no files')

            for path in paths:
                self.parse(path)

        self._path = old_path
        args.clear()

    def _match(self, option: str, args: List[str]) -> None:
        """Begin a conditional block"""

        # pylint: disable=unused-argument

        matching = True

        while args:
            match = args.pop(0).lower()

            if match[0] == '!':
                match = match[1:]
                negated = True
            else:
                negated = False

            if match == 'final' and self._final is None:
                self._final = False

            if match == 'all':
                result = True
            elif match == 'canonical':
                result = self._canonical
            elif match == 'final':
                result = cast(bool, self._final)
            else:
                match_val = self._match_val(match)

                if match != 'exec' and match_val is None:
                    self._error(f'Invalid match condition {match}')

                try:
                    arg = args.pop(0)
                except IndexError:
                    self._error(f'Missing {match} match pattern')

                if matching:
                    if match == 'exec':
                        result = _exec(arg)
                    elif match in ('address', 'localaddress'):
                        host_pat = HostPatternList(arg)
                        ip = ip_address(cast(str, match_val)) \
                            if match_val else None
                        result = host_pat.matches(None, match_val, ip)
                    else:
                        wild_pat = WildcardPatternList(arg)
                        result = wild_pat.matches(match_val)

            if matching and result == negated:
                matching = False

        self._matching = matching

    def _set_bool(self, option: str, args: List[str]) -> None:
        """Set a boolean config option"""

        value_str = args.pop(0).lower()

        if value_str in ('yes', 'true'):
            value = True
        elif value_str in ('no', 'false'):
            value = False
        else:
            self._error(f'Invalid {option} boolean value: {value_str}')

        if option not in self._options:
            self._options[option] = value

    def _set_bool_or_str(self, option: str, args: List[str]) -> None:
        """Set a boolean or string config option"""

        value_str = args.pop(0)
        value_lower = value_str.lower()

        if value_lower in ('yes', 'true'):
            value: Union[bool, str] = True
        elif value_lower in ('no', 'false'):
            value = False
        else:
            value = value_str

        if option not in self._options:
            self._options[option] = value

    def _set_int(self, option: str, args: List[str]) -> None:
        """Set an integer config option"""

        value_str = args.pop(0)

        try:
            value = int(value_str)
        except ValueError:
            self._error(f'Invalid {option} integer value: {value_str}')

        if option not in self._options:
            self._options[option] = value

    def _set_string(self, option: str, args: List[str]) -> None:
        """Set a string config option"""

        value_str = args.pop(0)

        if value_str.lower() == 'none':
            value = None
        else:
            value = value_str

        if option not in self._options:
            self._options[option] = value

    def _append_string(self, option: str, args: List[str]) -> None:
        """Append a string config option to a list"""

        value_str = args.pop(0)

        if value_str.lower() != 'none':
            if option in self._options:
                cast(List[str], self._options[option]).append(value_str)
            else:
                self._options[option] = [value_str]
        else:
            if option not in self._options:
                self._options[option] = []

    def _set_string_list(self, option: str, args: List[str]) -> None:
        """Set whitespace-separated string config options as a list"""

        if option not in self._options:
            if len(args) == 1 and args[0].lower() == 'none':
                self._options[option] = []
            else:
                self._options[option] = args[:]

        args.clear()

    def _append_string_list(self, option: str, args: List[str]) -> None:
        """Append whitespace-separated string config options to a list"""

        if option in self._options:
            cast(List[str], self._options[option]).extend(args)
        else:
            self._options[option] = args[:]

        args.clear()

    def _set_address_family(self, option: str, args: List[str]) -> None:
        """Set an address family config option"""

        value_str = args.pop(0).lower()

        if value_str == 'any':
            value = socket.AF_UNSPEC
        elif value_str == 'inet':
            value = socket.AF_INET
        elif value_str == 'inet6':
            value = socket.AF_INET6
        else:
            self._error(f'Invalid {option} value: {value_str}')

        if option not in self._options:
            self._options[option] = value

    def _set_canonicalize_host(self, option: str, args: List[str]) -> None:
        """Set a canonicalize host config option"""

        value_str = args.pop(0).lower()

        if value_str in ('yes', 'true'):
            value: Union[bool, str] = True
        elif value_str in ('no', 'false'):
            value = False
        elif value_str == 'always':
            value = value_str
        else:
            self._error(f'Invalid {option} value: {value_str}')

        if option not in self._options:
            self._options[option] = value

    def _set_rekey_limits(self, option: str, args: List[str]) -> None:
        """Set rekey limits config option"""

        byte_limit: Union[str, Tuple[()]] = args.pop(0).lower()

        if byte_limit == 'default':
            byte_limit = ()

        if args:
            time_limit: Optional[Union[str, Tuple[()]]] = args.pop(0).lower()

            if time_limit == 'none':
                time_limit = None
        else:
            time_limit = ()

        if option not in self._options:
            self._options[option] = byte_limit, time_limit

    def has_match_final(self) -> bool:
        """Return whether this config includes a 'Match final' block"""

        return self._final is not None

    def parse(self, path: Path) -> None:
        """Parse an OpenSSH config file and return matching declarations"""

        self._path = path
        self._line_no = 0
        self._matching = True
        self._tokens = {'%': '%'}

        logger.debug1('Reading config from "%s"', path)

        with open(path) as file:
            for line in file:
                self._line_no += 1

                line = line.strip()
                if not line or line[0] == '#':
                    continue

                try:
                    split_args = shlex.split(line)
                except ValueError as exc:
                    self._error(str(exc))

                args = []
                loption = ''
                allow_equal = True

                for i, arg in enumerate(split_args, 1):
                    if arg.startswith('='):
                        if len(arg) > 1:
                            args.append(arg[1:])
                    elif not allow_equal:
                        args.extend(split_args[i-1:])
                        break
                    elif arg.endswith('='):
                        args.append(arg[:-1])
                    elif '=' in arg:
                        arg, val = arg.split('=', 1)
                        args.append(arg)
                        args.append(val)
                    else:
                        args.append(arg)

                    if i == 1:
                        loption = args.pop(0).lower()
                        allow_equal = loption in self._conditionals

                if loption in self._no_split:
                    args = [line.lstrip()[len(loption):].strip()]

                if not self._matching and loption not in self._conditionals:
                    continue

                try:
                    option, handler = self._handlers[loption]
                except KeyError:
                    continue

                if not args:
                    self._error(f'Missing {option} value')

                handler(self, option, args)

                if args:
                    self._error(f'Extra data at end: {" ".join(args)}')

        self._set_tokens()

        for option in self._percent_expand:
            try:
                value = self._options[option]
            except KeyError:
                pass
            else:
                if isinstance(value, list):
                    value = [self._expand_val(item) for item in value]
                elif isinstance(value, str):
                    value = self._expand_val(value)

                self._options[option] = value

    def get_options(self, reload: bool) -> Dict[str, object]:
        """Return options to base a new config object on"""

        return self._last_options.copy() if reload else self._options.copy()

    @classmethod
    def load(cls, last_config: Optional['SSHConfig'],
             config_paths: ConfigPaths, reload: bool,
             canonical: bool, final: bool, *args: object) -> 'SSHConfig':
        """Load a list of OpenSSH config files into a config object"""

        config = cls(last_config, reload, canonical, final, *args)

        if config_paths:
            if isinstance(config_paths, (str, PurePath)):
                paths: Sequence[FilePath] = [config_paths]
            else:
                paths = config_paths

            for path in paths:
                config.parse(Path(path))

            config.loaded = True

        return config

    def get(self, option: str, default: object = None) -> object:
        """Get the value of a config option"""

        return self._options.get(option, default)

    def get_compression_algs(self) -> DefTuple[str]:
        """Return the compression algorithms to use"""

        compression = self.get('Compression')

        if compression is None:
            return ()
        elif compression:
            return 'zlib@openssh.com,zlib,none'
        else:
            return 'none,zlib@openssh.com,zlib'


class SSHClientConfig(SSHConfig):
    """Settings from an OpenSSH client config file"""

    _conditionals = {'host', 'match'}
    _no_split = {'proxycommand', 'remotecommand'}
    _percent_expand = {'CertificateFile', 'ForwardAgent', 'IdentityAgent',
                       'IdentityFile', 'ProxyCommand', 'RemoteCommand'}

    def __init__(self, last_config: 'SSHConfig', reload: bool,
                 canonical: bool, final: bool, local_user: str,
                 user: str, host: str, port: int) -> None:
        super().__init__(last_config, reload, canonical, final)

        self._local_user = local_user
        self._orig_host = host

        if user != ():
            self._options['User'] = user

        if port != ():
            self._options['Port'] = port

    def _match_val(self, match: str) -> object:
        """Return the value to match against in a match condition"""

        if match == 'host':
            return self._options.get('Hostname', self._orig_host)
        elif match == 'originalhost':
            return self._orig_host
        elif match == 'localuser':
            return self._local_user
        elif match == 'user':
            return self._options.get('User', self._local_user)
        elif match == 'tagged':
            return self._options.get('Tag', '')
        else:
            return None

    def _match_host(self, option: str, args: List[str]) -> None:
        """Begin a conditional block matching on host"""

        # pylint: disable=unused-argument

        pattern = ','.join(args)
        self._matching = WildcardPatternList(pattern).matches(self._orig_host)
        args.clear()

    def _set_hostname(self, option: str, args: List[str]) -> None:
        """Set hostname config option"""

        value = args.pop(0)

        if option not in self._options:
            self._tokens['h'] = \
                cast(str, self._options.get(option, self._orig_host))
            self._options[option] = self._expand_val(value)

    def _set_request_tty(self, option: str, args: List[str]) -> None:
        """Set a pseudo-terminal request config option"""

        value_str = args.pop(0).lower()

        if value_str in ('yes', 'true'):
            value: Union[bool, str] = True
        elif value_str in ('no', 'false'):
            value = False
        elif value_str in ('force', 'auto'):
            value = value_str
        else:
            self._error(f'Invalid {option} value: {value_str}')

        if option not in self._options:
            self._options[option] = value

    def _set_tokens(self) -> None:
        """Set the tokens available for percent expansion"""

        local_host = socket.gethostname()

        idx = local_host.find('.')
        short_local_host = local_host if idx < 0 else local_host[:idx]

        host = cast(str, self._options.get('Hostname', self._orig_host))
        port = str(self._options.get('Port', DEFAULT_PORT))
        user = cast(str, self._options.get('User') or self._local_user)
        home = os.path.expanduser('~')

        conn_info = ''.join((local_host, host, port, user))
        conn_hash = sha1(conn_info.encode('utf-8')).hexdigest()

        self._tokens.update({'C': conn_hash,
                             'h': host,
                             'L': short_local_host,
                             'l': local_host,
                             'n': self._orig_host,
                             'p': port,
                             'r': user,
                             'u': self._local_user})

        if home != '~':
            self._tokens['d'] = home

        if hasattr(os, 'getuid'):
            self._tokens['i'] = str(os.getuid())

    _handlers = {option.lower(): (option, handler) for option, handler in (
        ('Host',                            _match_host),
        ('Match',                           SSHConfig._match),
        ('Include',                         SSHConfig._include),

        ('AddressFamily',                   SSHConfig._set_address_family),
        ('BindAddress',                     SSHConfig._set_string),
        ('CanonicalDomains',                SSHConfig._set_string_list),
        ('CanonicalizeFallbackLocal',       SSHConfig._set_bool),
        ('CanonicalizeHostname',            SSHConfig._set_canonicalize_host),
        ('CanonicalizeMaxDots',             SSHConfig._set_int),
        ('CanonicalizePermittedCNAMEs',     SSHConfig._set_string_list),
        ('CASignatureAlgorithms',           SSHConfig._set_string),
        ('CertificateFile',                 SSHConfig._append_string),
        ('ChallengeResponseAuthentication', SSHConfig._set_bool),
        ('Ciphers',                         SSHConfig._set_string),
        ('Compression',                     SSHConfig._set_bool),
        ('ConnectTimeout',                  SSHConfig._set_int),
        ('EnableSSHKeySign',                SSHConfig._set_bool),
        ('ForwardAgent',                    SSHConfig._set_bool_or_str),
        ('ForwardX11Trusted',               SSHConfig._set_bool),
        ('GlobalKnownHostsFile',            SSHConfig._set_string_list),
        ('GSSAPIAuthentication',            SSHConfig._set_bool),
        ('GSSAPIDelegateCredentials',       SSHConfig._set_bool),
        ('GSSAPIKeyExchange',               SSHConfig._set_bool),
        ('HostbasedAuthentication',         SSHConfig._set_bool),
        ('HostKeyAlgorithms',               SSHConfig._set_string),
        ('Hostname',                        _set_hostname),
        ('HostKeyAlias',                    SSHConfig._set_string),
        ('IdentitiesOnly',                  SSHConfig._set_bool),
        ('IdentityAgent',                   SSHConfig._set_string),
        ('IdentityFile',                    SSHConfig._append_string),
        ('KbdInteractiveAuthentication',    SSHConfig._set_bool),
        ('KexAlgorithms',                   SSHConfig._set_string),
        ('MACs',                            SSHConfig._set_string),
        ('PasswordAuthentication',          SSHConfig._set_bool),
        ('PKCS11Provider',                  SSHConfig._set_string),
        ('PreferredAuthentications',        SSHConfig._set_string),
        ('Port',                            SSHConfig._set_int),
        ('ProxyCommand',                    SSHConfig._set_string),
        ('ProxyJump',                       SSHConfig._set_string),
        ('PubkeyAuthentication',            SSHConfig._set_bool),
        ('RekeyLimit',                      SSHConfig._set_rekey_limits),
        ('RemoteCommand',                   SSHConfig._set_string),
        ('RequestTTY',                      _set_request_tty),
        ('SendEnv',                         SSHConfig._append_string_list),
        ('ServerAliveCountMax',             SSHConfig._set_int),
        ('ServerAliveInterval',             SSHConfig._set_int),
        ('SetEnv',                          SSHConfig._set_string_list),
        ('Tag',                             SSHConfig._set_string),
        ('TCPKeepAlive',                    SSHConfig._set_bool),
        ('User',                            SSHConfig._set_string),
        ('UserKnownHostsFile',              SSHConfig._set_string_list)
    )}


class SSHServerConfig(SSHConfig):
    """Settings from an OpenSSH server config file"""

    def __init__(self, last_config: 'SSHConfig', reload: bool,
                 canonical: bool, final: bool, local_addr: str,
                 local_port: int, user: str, host: str, addr: str) -> None:
        super().__init__(last_config, reload, canonical, final)

        self._local_addr = local_addr
        self._local_port = local_port
        self._user = user
        self._host = host or addr
        self._addr = addr

    def _match_val(self, match: str) -> object:
        """Return the value to match against in a match condition"""

        if match == 'localaddress':
            return self._local_addr
        elif match == 'localport':
            return str(self._local_port)
        elif match == 'user':
            return self._user
        elif match == 'host':
            return self._host
        elif match == 'address':
            return self._addr
        else:
            return None

    def _set_tokens(self) -> None:
        """Set the tokens available for percent expansion"""

        self._tokens.update({'u': self._user})

    _handlers = {option.lower(): (option, handler) for option, handler in (
        ('Match',                           SSHConfig._match),
        ('Include',                         SSHConfig._include),

        ('AddressFamily',                   SSHConfig._set_address_family),
        ('AuthorizedKeysFile',              SSHConfig._set_string_list),
        ('AllowAgentForwarding',            SSHConfig._set_bool),
        ('BindAddress',                     SSHConfig._set_string),
        ('CanonicalDomains',                SSHConfig._set_string_list),
        ('CanonicalizeFallbackLocal',       SSHConfig._set_bool),
        ('CanonicalizeHostname',            SSHConfig._set_canonicalize_host),
        ('CanonicalizeMaxDots',             SSHConfig._set_int),
        ('CanonicalizePermittedCNAMEs',     SSHConfig._set_string_list),
        ('CASignatureAlgorithms',           SSHConfig._set_string),
        ('ChallengeResponseAuthentication', SSHConfig._set_bool),
        ('Ciphers',                         SSHConfig._set_string),
        ('ClientAliveCountMax',             SSHConfig._set_int),
        ('ClientAliveInterval',             SSHConfig._set_int),
        ('Compression',                     SSHConfig._set_bool),
        ('GSSAPIAuthentication',            SSHConfig._set_bool),
        ('GSSAPIKeyExchange',               SSHConfig._set_bool),
        ('HostbasedAuthentication',         SSHConfig._set_bool),
        ('HostCertificate',                 SSHConfig._append_string),
        ('HostKey',                         SSHConfig._append_string),
        ('KbdInteractiveAuthentication',    SSHConfig._set_bool),
        ('KexAlgorithms',                   SSHConfig._set_string),
        ('LoginGraceTime',                  SSHConfig._set_int),
        ('MACs',                            SSHConfig._set_string),
        ('PasswordAuthentication',          SSHConfig._set_bool),
        ('PermitTTY',                       SSHConfig._set_bool),
        ('Port',                            SSHConfig._set_int),
        ('PubkeyAuthentication',            SSHConfig._set_bool),
        ('RekeyLimit',                      SSHConfig._set_rekey_limits),
        ('TCPKeepAlive',                    SSHConfig._set_bool),
        ('UseDNS',                          SSHConfig._set_bool)
    )}



# Copyright (c) 2013-2024 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""AsyncSSH version information"""

__author__ = 'Ron Frederick'

__author_email__ = 'ronf@timeheart.net'

__url__ = 'http://asyncssh.timeheart.net'

__version__ = '2.21.0'



# Copyright (c) 2015-2021 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""Pattern matching for principal and host names"""

from fnmatch import fnmatch
from typing import Union

from .misc import IPAddress, ip_network


_HostPattern = Union['WildcardHostPattern', 'CIDRHostPattern']
_AnyPattern = Union['WildcardPattern', _HostPattern]


class _BaseWildcardPattern:
    """A base class for matching '*' and '?' wildcards"""

    def __init__(self, pattern: str):
        # We need to escape square brackets in host patterns if we
        # want to use Python's fnmatch.
        self._pattern = ''.join('[[]' if ch == '[' else
                                '[]]' if ch == ']' else
                                ch for ch in pattern)

    def _matches(self, value: str) -> bool:
        """Return whether a wild card pattern matches a value"""

        return fnmatch(value, self._pattern)


class WildcardPattern(_BaseWildcardPattern):
    """A pattern matcher for '*' and '?' wildcards"""

    def matches(self, value: str) -> bool:
        """Return whether a wild card pattern matches a value"""

        return super()._matches(value)


class WildcardHostPattern(_BaseWildcardPattern):
    """Match a host name or address against a wildcard pattern"""

    def matches(self, host: str, addr: str, _ip: IPAddress) -> bool:
        """Return whether a host or address matches a wild card host pattern"""

        return (bool(host) and super()._matches(host)) or \
               (bool(addr) and super()._matches(addr))


class CIDRHostPattern:
    """Match IPv4/v6 address against CIDR-style subnet pattern"""

    def __init__(self, pattern: str):
        self._network = ip_network(pattern)

    def matches(self, _host: str, _addr: str, ip: IPAddress) -> bool:
        """Return whether an IP address matches a CIDR address pattern"""

        return bool(ip) and ip in self._network


class _PatternList:
    """Match against a list of comma-separated positive and negative patterns

       This class is a base class for building a pattern matcher that
       takes a set of comma-separated positive and negative patterns,
       returning `True` if one or more positive patterns match and
       no negative ones do.

       The pattern matching is done by objects returned by the
       build_pattern method. The arguments passed in when a match
       is performed will vary depending on what class build_pattern
       returns.

    """

    def __init__(self, patterns: str):
        self._pos_patterns = []
        self._neg_patterns = []

        for pattern in patterns.split(','):
            if pattern.startswith('!'):
                negate = True
                pattern = pattern[1:]
            else:
                negate = False

            matcher = self.build_pattern(pattern)

            if negate:
                self._neg_patterns.append(matcher)
            else:
                self._pos_patterns.append(matcher)

    def build_pattern(self, pattern: str) -> _AnyPattern:
        """Abstract method to build a pattern object"""

        raise NotImplementedError

    def matches(self, *args) -> bool:
        """Match a set of values against positive & negative pattern lists"""

        pos_match = any(p.matches(*args) for p in self._pos_patterns)
        neg_match = any(p.matches(*args) for p in self._neg_patterns)

        return pos_match and not neg_match


class WildcardPatternList(_PatternList):
    """Match names against wildcard patterns"""

    def build_pattern(self, pattern: str) -> WildcardPattern:
        """Build a wild card pattern"""

        return WildcardPattern(pattern)


class HostPatternList(_PatternList):
    """Match host names & addresses against wildcard and CIDR patterns"""

    def build_pattern(self, pattern: str) -> _HostPattern:
        """Build a CIDR address or wild card host pattern"""

        try:
            return CIDRHostPattern(pattern)
        except ValueError:
            return WildcardHostPattern(pattern)



# Copyright (c) 2015-2021 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""Chacha20-Poly1305 symmetric encryption handler"""

from ctypes import c_ulonglong, create_string_buffer
from typing import Optional, Tuple

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends.openssl import backend
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import ChaCha20
from cryptography.hazmat.primitives.poly1305 import Poly1305

from .cipher import register_cipher


if backend.poly1305_supported():
    _CTR_0 = (0).to_bytes(8, 'little')
    _CTR_1 = (1).to_bytes(8, 'little')

    _POLY1305_KEYBYTES = 32

    def chacha20(key: bytes, data: bytes, nonce: bytes, ctr: int) -> bytes:
        """Encrypt/decrypt a block of data with the ChaCha20 cipher"""

        return Cipher(ChaCha20(key, (_CTR_1 if ctr else _CTR_0) + nonce),
                      mode=None).encryptor().update(data)

    def poly1305_key(key: bytes, nonce: bytes) -> bytes:
        """Derive a Poly1305 key"""

        return chacha20(key, _POLY1305_KEYBYTES * b'\0', nonce, 0)

    def poly1305(key: bytes, data: bytes, nonce: bytes) -> bytes:
        """Compute a Poly1305 tag for a block of data"""

        return Poly1305.generate_tag(poly1305_key(key, nonce), data)

    def poly1305_verify(key: bytes, data: bytes,
                        nonce: bytes, tag: bytes) -> bool:
        """Verify a Poly1305 tag for a block of data"""

        try:
            Poly1305.verify_tag(poly1305_key(key, nonce), data, tag)
            return True
        except InvalidSignature:
            return False

    chacha_available = True
else: # pragma: no cover
    try:
        from libnacl import nacl

        _chacha20 = nacl.crypto_stream_chacha20
        _chacha20_xor_ic = nacl.crypto_stream_chacha20_xor_ic

        _POLY1305_BYTES = nacl.crypto_onetimeauth_poly1305_bytes()
        _POLY1305_KEYBYTES = nacl.crypto_onetimeauth_poly1305_keybytes()

        _poly1305 = nacl.crypto_onetimeauth_poly1305
        _poly1305_verify = nacl.crypto_onetimeauth_poly1305_verify

        def chacha20(key: bytes, data: bytes, nonce: bytes, ctr: int) -> bytes:
            """Encrypt/decrypt a block of data with the ChaCha20 cipher"""

            datalen = len(data)
            result = create_string_buffer(datalen)
            ull_datalen = c_ulonglong(datalen)
            ull_ctr = c_ulonglong(ctr)

            _chacha20_xor_ic(result, data, ull_datalen, nonce, ull_ctr, key)

            return result.raw

        def poly1305_key(key: bytes, nonce: bytes) -> bytes:
            """Derive a Poly1305 key"""

            polykey = create_string_buffer(_POLY1305_KEYBYTES)
            ull_polykeylen = c_ulonglong(_POLY1305_KEYBYTES)

            _chacha20(polykey, ull_polykeylen, nonce, key)

            return polykey.raw

        def poly1305(key: bytes, data: bytes, nonce: bytes) -> bytes:
            """Compute a Poly1305 tag for a block of data"""

            tag = create_string_buffer(_POLY1305_BYTES)
            ull_datalen = c_ulonglong(len(data))
            polykey = poly1305_key(key, nonce)

            _poly1305(tag, data, ull_datalen, polykey)

            return tag.raw

        def poly1305_verify(key: bytes, data: bytes,
                            nonce: bytes, tag: bytes) -> bool:
            """Verify a Poly1305 tag for a block of data"""

            ull_datalen = c_ulonglong(len(data))
            polykey = poly1305_key(key, nonce)

            return _poly1305_verify(tag, data, ull_datalen, polykey) == 0

        chacha_available = True
    except (ImportError, OSError, AttributeError):
        chacha_available = False


class ChachaCipher:
    """Shim for Chacha20-Poly1305 symmetric encryption"""

    def __init__(self, key: bytes):
        keylen = len(key) // 2
        self._key = key[:keylen]
        self._adkey = key[keylen:]

    def encrypt_and_sign(self, header: bytes, data: bytes,
                         nonce: bytes) -> Tuple[bytes, bytes]:
        """Encrypt and sign a block of data"""

        header = chacha20(self._adkey, header, nonce, 0)
        data = chacha20(self._key, data, nonce, 1)
        tag = poly1305(self._key, header + data, nonce)

        return header + data, tag

    def decrypt_header(self, header: bytes, nonce: bytes) -> bytes:
        """Decrypt header data"""

        return chacha20(self._adkey, header, nonce, 0)

    def verify_and_decrypt(self, header: bytes, data: bytes,
                           nonce: bytes, tag: bytes) -> Optional[bytes]:
        """Verify the signature of and decrypt a block of data"""

        if poly1305_verify(self._key, header + data, nonce, tag):
            return chacha20(self._key, data, nonce, 1)
        else:
            return None


if chacha_available: # pragma: no branch
    register_cipher('chacha20-poly1305', 64, 0, 1)



# Copyright (c) 2019-2021 by Ron Frederick <ronf@timeheart.net> and others.
#
# This program and the accompanying materials are made available under
# the terms of the Eclipse Public License v2.0 which accompanies this
# distribution and is available at:
#
#     http://www.eclipse.org/legal/epl-2.0/
#
# This program may also be made available under the following secondary
# licenses when the conditions for such availability set forth in the
# Eclipse Public License v2.0 are satisfied:
#
#    GNU General Public License, Version 2.0, or any later versions of
#    that license
#
# SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
#
# Contributors:
#     Ron Frederick - initial implementation, API, and documentation

"""EdDSA public key encryption handler"""

from typing import Optional, Tuple, Union, cast

from .asn1 import ASN1DecodeError, ObjectIdentifier, der_encode, der_decode
from .crypto import EdDSAPrivateKey, EdDSAPublicKey
from .crypto import ed25519_available, ed448_available
from .packet import String, SSHPacket
from .public_key import OMIT, SSHKey, SSHOpenSSHCertificateV01
from .public_key import KeyImportError, KeyExportError
from .public_key import register_public_key_alg, register_certificate_alg
from .public_key import register_x509_certificate_alg


_PrivateKeyArgs = Tuple[bytes]
_PublicKeyArgs = Tuple[bytes]


class _EdKey(SSHKey):
    """Handler for EdDSA public key encryption"""

    _key: Union[EdDSAPrivateKey, EdDSAPublicKey]

    algorithm = b''

    def __eq__(self, other: object) -> bool:
        # This isn't protected access - both objects are _EdKey instances
        # pylint: disable=protected-access

        return (isinstance(other, type(self)) and
                self._key.public_value == other._key.public_value and
                self._key.private_value == other._key.private_value)

    def __hash__(self) -> int:
        return hash((self._key.public_value, self._key.private_value))

    @classmethod
    def generate(cls, algorithm: bytes) -> '_EdKey': # type: ignore
        """Generate a new EdDSA private key"""

        # pylint: disable=arguments-differ

        # Strip 'ssh-' prefix of algorithm to get curve_id
        return cls(EdDSAPrivateKey.generate(algorithm[4:]))

    @classmethod
    def make_private(cls, key_params: object) -> SSHKey:
        """Construct an EdDSA private key"""

        try:
            private_value, = cast(_PrivateKeyArgs, key_params)

            return cls(EdDSAPrivateKey.construct(cls.algorithm[4:],
                                                 private_value))
        except (TypeError, ValueError):
            raise KeyImportError('Invalid EdDSA private key') from None

    @classmethod
    def make_public(cls, key_params: object) -> SSHKey:
        """Construct an EdDSA public key"""

        try:
            public_value, = cast(_PublicKeyArgs, key_params)

            return cls(EdDSAPublicKey.construct(cls.algorithm[4:],
                                                public_value))
        except (TypeError, ValueError):
            raise KeyImportError('Invalid EdDSA public key') from None

    @classmethod
    def decode_pkcs8_private(cls, alg_params: object,
                             data: bytes) -> Optional[_PrivateKeyArgs]:
        """Decode a PKCS#8 format EdDSA private key"""

        # pylint: disable=unused-argument

        try:
            return (cast(bytes, der_decode(data)),)
        except ASN1DecodeError:
            return None

    @classmethod
    def decode_pkcs8_public(cls, alg_params: object,
                            data: bytes) -> Optional[_PublicKeyArgs]:
        """Decode a PKCS#8 format EdDSA public key"""

        # pylint: disable=unused-argument

        return (data,)

    @classmethod
    def decode_ssh_private(cls, packet: SSHPacket) -> _PrivateKeyArgs:
        """Decode an SSH format EdDSA private key"""

        public_value = packet.get_string()
        private_value = packet.get_string()

        return (private_value[:-len(public_value)],)

    @classmethod
    def decode_ssh_public(cls, packet: SSHPacket) -> _PublicKeyArgs:
        """Decode an SSH format EdDSA public key"""

        public_value = packet.get_string()

        return (public_value,)

    def encode_pkcs8_private(self) -> Tuple[object, object]:
        """Encode a PKCS#8 format EdDSA private key"""

        if not self._key.private_value:
            raise KeyExportError('Key is not private')

        return OMIT, der_encode(self._key.private_value)

    def encode_pkcs8_public(self) -> Tuple[object, object]:
        """Encode a PKCS#8 format EdDSA public key"""

        return OMIT, self._key.public_value

    def encode_ssh_private(self) -> bytes:
        """Encode an SSH format EdDSA private key"""

        if self._key.private_value is None:
            raise KeyExportError('Key is not private')

        return b''.join((String(self._key.public_value),
                         String(self._key.private_value +
                                self._key.public_value)))

    def encode_ssh_public(self) -> bytes:
        """Encode an SSH format EdDSA public key"""

        return String(self._key.public_value)

    def encode_agent_cert_private(self) -> bytes:
        """Encode EdDSA certificate private key data for agent"""

        return self.encode_ssh_private()

    def sign_ssh(self, data: bytes, sig_algorithm: bytes) -> bytes:
        """Compute an SSH-encoded signature of the specified data"""

        # pylint: disable=unused-argument

        if not self._key.private_value:
            raise ValueError('Private key needed for signing')

        return String(self._key.sign(data))

    def verify_ssh(self, data: bytes, sig_algorithm: bytes,
                   packet: SSHPacket) -> bool:
        """Verify an SSH-encoded signature of the specified data"""

        # pylint: disable=unused-argument

        sig = packet.get_string()
        packet.check_end()

        return self._key.verify(data, sig)


class _Ed25519Key(_EdKey):
    """Handler for Curve25519 public key encryption"""

    algorithm = b'ssh-ed25519'
    pkcs8_oid = ObjectIdentifier('1.3.101.112')
    sig_algorithms = (algorithm,)
    x509_algorithms = (b'x509v3-' + algorithm,)
    all_sig_algorithms = set(sig_algorithms)


class _Ed448Key(_EdKey):
    """Handler for Curve448 public key encryption"""

    algorithm = b'ssh-ed448'
    pkcs8_oid = ObjectIdentifier('1.3.101.113')
    sig_algorithms = (algorithm,)
    x509_algorithms = (b'x509v3-' + algorithm,)
    all_sig_algorithms = set(sig_algorithms)


if ed25519_available: # pragma: no branch
    register_public_key_alg(b'ssh-ed25519', _Ed25519Key, True)

    register_certificate_alg(1, b'ssh-ed25519',
                             b'ssh-ed25519-cert-v01@openssh.com',
                             _Ed25519Key, SSHOpenSSHCertificateV01, True)

    for alg in _Ed25519Key.x509_algorithms:
        register_x509_certificate_alg(alg, True)

if ed448_available: # pragma: no branch
    register_public_key_alg(b'ssh-ed448', _Ed448Key, True)

    register_certificate_alg(1, b'ssh-ed448', b'ssh-ed448-cert-v01@openssh.com',
                             _Ed448Key, SSHOpenSSHCertificateV01, True)

    for alg in _Ed448Key.x509_algorithms:
        register_x509_certificate_alg(alg, True)
